import inspect
import os
import os.path

import mock
import pytest

from batou import remote_core


@pytest.fixture(autouse=True)
def reset_global_vars():
    yield
    remote_core.target_directory = None
    remote_core.environment = None
    remote_core.deployment = None
    remote_core.channel = None


def test_deployment_and_channel_exist_as_names():
    assert remote_core.channel is None
    assert remote_core.deployment is None


def test_lock():
    remote_core.lock()


def test_cmd():
    result = remote_core.cmd('echo "asdf"')
    assert result == (b"asdf\n", b"")


@pytest.fixture
def mock_remote_core(monkeypatch):
    # Suppress active actions in the remote_core module
    monkeypatch.setattr(remote_core, "cmd", mock.Mock())


def test_update_code_existing_target(mock_remote_core, tmpdir):
    remote_core.cmd.side_effect = [
        ("", ""),
        ("", ""),
        ("", ""),
        ('[{"id": "12345"}]', ""),
    ]

    remote_core.ensure_repository(str(tmpdir), "hg-pull")
    remote_core.hg_pull_code("http://bitbucket.org/flyingcircus/batou")
    assert "12345" == remote_core.hg_update_working_copy("default")

    calls = iter(x[1][0] for x in remote_core.cmd.mock_calls)
    assert next(calls).startswith("hg init /")
    assert next(calls) == "hg pull http://bitbucket.org/flyingcircus/batou"
    assert next(calls) == "hg up -C default"
    assert next(calls) == "hg id -Tjson"
    assert remote_core.cmd.call_count == 4


def test_update_code_new_target(mock_remote_core, tmpdir):
    remote_core.cmd.side_effect = [
        ("", ""),
        ("", ""),
        ("", ""),
        ('[{"id": "12345"}]', ""),
    ]
    remote_core.ensure_repository(str(tmpdir) + "/foo", "hg-bundle")
    remote_core.hg_pull_code("http://bitbucket.org/flyingcircus/batou")
    assert "12345" == remote_core.hg_update_working_copy("default")

    assert remote_core.cmd.call_count == 4
    calls = iter(x[1][0] for x in remote_core.cmd.mock_calls)
    assert next(calls) == "hg init {}".format(remote_core.target_directory)
    assert next(calls) == "hg pull http://bitbucket.org/flyingcircus/batou"
    assert next(calls) == "hg up -C default"
    assert next(calls) == "hg id -Tjson"


def test_hg_bundle_shipping(mock_remote_core, tmpdir):
    remote_core.ensure_repository(str(tmpdir) + "/foo", "hg-bundle")
    remote_core.cmd.side_effect = [
        (b'[{"id": "41fea38ce5d3"}]', None),
        (b'[{"node": "revision-a"},{"node": "revision-b"}]', None),
        (b"", b""),
        (b"", b""),
        (b'[{"id": "revision-b"}]', b""),
    ]
    heads = remote_core.hg_current_heads()
    assert heads == ["revision-a", "revision-b"]
    remote_core.hg_unbundle_code()
    assert remote_core.hg_update_working_copy("default") == "revision-b"

    assert remote_core.cmd.call_count == 6
    calls = iter(x[1][0] for x in remote_core.cmd.mock_calls)
    assert next(calls) == "hg init {}".format(remote_core.target_directory)
    assert next(calls) == "hg id -Tjson"
    assert next(calls) == "hg heads -Tjson"
    assert next(calls) == "hg -y unbundle batou-bundle.hg"
    assert next(calls) == "hg up -C default"
    assert next(calls) == "hg id -Tjson"


def test_build_batou_fresh_install(mock_remote_core, tmpdir):
    remote_core.ensure_repository(str(tmpdir), "hg-pull")
    remote_core.ensure_base("asdf")
    remote_core.cmd.reset_mock()
    remote_core.build_batou()
    calls = iter([x[1][0] for x in remote_core.cmd.mock_calls])
    assert remote_core.cmd.call_count == 1
    assert next(calls) == "./batou --help"


def test_build_batou_virtualenv_exists(mock_remote_core, tmpdir):
    remote_core.ensure_repository(str(tmpdir), "hg-pull")
    remote_core.ensure_base("asdf")
    os.mkdir(remote_core.target_directory + "/bin")
    with open(remote_core.target_directory + "/bin/python3", "w"):
        # Break python
        pass
    remote_core.build_batou()
    calls = iter([x[1][0] for x in remote_core.cmd.mock_calls])
    assert remote_core.cmd.call_count == 2
    next(calls)  # skip ensure_repository
    assert next(calls) == "./batou --help"


def test_expand_deployment_base(tmpdir):
    with mock.patch("os.path.expanduser") as expanduser:
        expanduser.return_value = str(tmpdir)
        remote_core.ensure_repository("~/deployment", "rsync")
    assert remote_core.target_directory == str(tmpdir)


def test_deploy_component(monkeypatch):
    import batou

    monkeypatch.setattr(remote_core, "deployment", mock.Mock())
    monkeypatch.setattr(batou, "output", mock.Mock())
    remote_core.deploy("foo")
    assert remote_core.deployment.deploy.call_count == 1


def test_whoami():
    # Smoke test
    assert remote_core.whoami() != ""


class DummyChannel(object):

    _closed = False

    def __init__(self):
        self.receivequeue = []
        self.sendqueue = []

    def isclosed(self):
        return self._closed

    def receive(self, timeout=None):
        result = self.receivequeue.pop(0)
        if not self.receivequeue:
            self._closed = True
        return result

    def send(self, item):
        self.sendqueue.append(item)


@pytest.fixture
def remote_core_mod():
    channel = DummyChannel()

    local_namespace = dict(channel=channel, __name__="__channelexec__")

    remote_core_mod = compile(
        inspect.getsource(remote_core),
        inspect.getsourcefile(remote_core),
        "exec",
    )

    def run():
        exec(remote_core_mod, local_namespace)

    return (channel, run)


def test_channelexec_already_closed(remote_core_mod):
    channel, run = remote_core_mod
    channel._closed = True
    run()
    assert channel.receivequeue == []
    assert channel.sendqueue == []


def test_channelexec_echo_cmd(remote_core_mod):
    channel, run = remote_core_mod
    channel.receivequeue.append(("cmd", ('echo "asdf"',), {}))
    run()
    assert channel.isclosed()
    assert channel.receivequeue == []
    assert channel.sendqueue == [("batou-result", (b"asdf\n", b""))]


def test_channelexec_multiple_echo_cmds(remote_core_mod):
    channel, run = remote_core_mod
    channel.receivequeue.append(("cmd", ('echo "asdf1"',), {}))
    channel.receivequeue.append(("cmd", ('echo "asdf2"',), {}))
    run()
    assert channel.isclosed()
    assert channel.receivequeue == []
    assert channel.sendqueue == [
        ("batou-result", (b"asdf1\n", b"")),
        ("batou-result", (b"asdf2\n", b"")),
    ]


def test_channelexec_handle_exception(remote_core_mod):
    channel, run = remote_core_mod
    channel.receivequeue.append(("cmd", ("fdjkahfkjdasbfda",), {}))
    run()
    assert channel.isclosed()
    assert channel.receivequeue == []
    response = iter(channel.sendqueue)

    assert next(response) == (
        "batou-output",
        "line",
        ("ERROR: fdjkahfkjdasbfda",),
        {"bold": True, "red": True},
    )

    assert next(response) == (
        "batou-output",
        "line",
        ("    Return code: 127",),
        {"red": True},
    )

    assert next(response) == (
        "batou-output",
        "line",
        ("STDOUT",),
        {"red": True},
    )

    assert next(response) == ("batou-output", "line", ("",), {})

    assert next(response) == (
        "batou-output",
        "line",
        ("STDERR",),
        {"red": True},
    )

    # Different /bin/sh versions have different error reporting
    r_command_not_found = next(response)
    assert "fdjkahfkjdasbfda" in r_command_not_found[2][0]
    assert "not found" in r_command_not_found[2][0]

    assert next(response) == ("batou-error", None)

    with pytest.raises(StopIteration):
        next(response)


def test_git_remote_init_bundle(tmpdir):
    source = tmpdir.mkdir("source")
    dest = tmpdir.mkdir("dest")
    with source.as_cwd():
        remote_core.cmd("git init")
        source.join("foo.txt").write("bar")
        remote_core.cmd("git add foo.txt")
        remote_core.cmd("git commit -m bar")
        remote_core.cmd(
            "git bundle create {} master ".format(dest.join("batou-bundle.git"))
        )

    remote_core.ensure_repository(str(dest), "git-bundle")
    remote_core.git_unbundle_code()
    remote_core.git_update_working_copy("master")

    assert "bar" == dest.join("foo.txt").read()


def test_git_remote_init_pull(tmpdir):
    source = tmpdir.mkdir("source")
    dest = tmpdir.mkdir("dest")
    with source.as_cwd():
        remote_core.cmd("git init")
        source.join("foo.txt").write("bar")
        remote_core.cmd("git add foo.txt")
        remote_core.cmd("git commit -m bar")

        remote_core.ensure_repository(str(dest), "git-bundle")
        remote_core.git_pull_code(str(source), "master")
        remote_core.git_update_working_copy("master")

    assert "bar" == dest.join("foo.txt").read()


def test_Deployment_sets_os_environ_on_load(monkeypatch):
    """If ``os_env`` is given to ``Deployment`` it changes os.environ, ...or

    when the ``load()`` method is called.
    """
    monkeypatch.chdir("examples/tutorial-helloworld")

    assert "MY_ENV_VAR" not in os.environ

    dep = remote_core.Deployment(
        env_name="tutorial",
        host_name="localhost",
        overrides={},
        secret_files=None,
        secret_data=None,
        host_data={},
        timeout=None,
        platform=None,
        os_env={"MY_ENV_VAR": "MY-VALUE"},
    )
    dep.load()

    assert os.environ["MY_ENV_VAR"] == "MY-VALUE"



import difflib
import glob
import grp
import itertools
import json
import os.path
import pwd
import re
import shutil
import stat
import tempfile

import yaml

import batou
from batou import output
from batou.component import Attribute, Component
from batou.utils import dict_merge

# Explain the logic that in a multi-user and concurrent system there isn't really a coarse grained guarantee around files not existing.

# What this can guarantee, however, is that during the lifetime of this function call, there will have been an atomic moment where the file did not exist, or, if it existed and can not be removed, it will raise an error.

# This means: if the file wasn't there in the first place, we're fine. If the file existed and we removed it, we are fine, we don't care whether someone else might have recreated it immediately after. If the file existed and was removed by someone else, we don't care either. If the file existed and we can't remove it, we need to fail.


def ensure_path_nonexistent(path: str) -> None:
    """Ensure that the given path does not exist.

    In a multi-user/concurrent environment, it is possible that a path is created
    at any time. This function will ensure that the path does not exist at some
    point in time during the execution of this function.

    During the lifetime of this function call, there will have been an atomic
    moment where the file did not exist, or, if it existed and can not be
    removed, it will raise an error.

    This function will not guarantee that the path does not exist after this
    function has returned.
    """

    if not os.path.lexists(path):
        return

    parent_dir = os.path.dirname(os.path.abspath(path))
    path_basename = os.path.basename(os.path.normpath(path))

    with tempfile.TemporaryDirectory(dir=parent_dir) as temp_dir:
        try:
            os.rename(path, os.path.join(temp_dir, path_basename))
        except FileNotFoundError as e:
            # The file was not there in the first place, we're done.
            pass


class File(Component):

    namevar = "path"

    ensure = "file"  # or: directory, symlink

    # Content oriented parameters
    content = None
    source = ""
    is_template = True
    template_context = None
    template_args = None  # dict, actually
    encoding = "utf-8"

    # Unix attributes
    owner = None
    group = None
    mode = None

    # Symlink parameters
    link_to = ""

    # Leading directory creation
    leading = False

    # Signal that the content is sensitive data.
    sensitive_data = None

    def configure(self):
        self._unmapped_path = self.path
        self.path = self.map(self.path)
        if self.ensure == "file":
            self += Presence(self.path, leading=self.leading)
        elif self.ensure == "directory":
            self += Directory(
                self.path, leading=self.leading, source=self.source
            )
        elif self.ensure == "symlink":
            self += Symlink(self.path, source=self.link_to)
        else:
            raise ValueError(
                "Ensure must be one of: file, directory, "
                "symlink not %s" % self.ensure
            )
        # variation: content or source explicitly given

        # The mode needs to be set early to allow batou to get out of
        # accidental "permission denied" situations.
        if self.mode:
            self += Mode(self.path, mode=self.mode)

        # no content or source given but file with same name
        # exists
        if self.ensure == "file" and self.content is None and not self.source:
            guess_source = self.root.defdir + "/" + os.path.basename(self.path)
            if os.path.isfile(guess_source):
                self.source = guess_source
            else:
                # Avoid the edge case where we want to support a very simple
                # case: specify File('asdf') and have an identical named file
                # in the component definition directory that will be templated
                # to the work directory.
                #
                # However, if you mis-spell the file, then you might
                # accidentally end up with an empty file in the work directory.
                # If you really want an empty File then you can either use
                # Presence(), or (recommended) use File('asdf', content='') to
                # make this explicit. We don't want to accidentally confuse the
                # convenience case (template a simple file) and an edge case
                # (have an empty file)
                raise ValueError(
                    "Missing implicit template file {}. Or did you want "
                    "to create an empty file? Then use File('{}', content='').".format(
                        guess_source, self._unmapped_path
                    )
                )
        if self.ensure == "file" and (self.content or self.source):
            if self.template_args is None:
                self.template_args = dict()
            if not self.template_context:
                self.template_context = self.parent
            content = Content(
                self.path,
                source=self.source,
                is_template=self.is_template,
                template_context=self.template_context,
                template_args=self.template_args,
                encoding=self.encoding,
                content=self.content,
                sensitive_data=self.sensitive_data,
            )
            self += content
            self.content = content.content

        if self.owner:
            self += Owner(self.path, owner=self.owner)

        if self.group:
            self += Group(self.path, group=self.group)

    @property
    def namevar_for_breadcrumb(self):
        relpath = os.path.relpath(self.path, self.environment.base_dir)
        if not relpath.startswith(".."):
            return relpath
        return os.path.abspath(self.path)

    def last_updated(self, key="st_mtime"):
        if not os.path.exists(self.path):
            return None
        return getattr(os.stat(self.path), key)


class BinaryFile(File):

    is_template = False
    encoding = None


class Presence(Component):

    namevar = "path"
    leading = False

    def configure(self):
        self.path = self.map(self.path)
        if self.leading:
            self += Directory(os.path.dirname(self.path), leading=self.leading)

    def verify(self):
        assert os.path.isfile(self.path)

    def update(self):
        ensure_path_nonexistent(self.path)
        with open(self.path, "w"):
            # We're just touching it.
            pass

    @property
    def namevar_for_breadcrumb(self):
        if isinstance(self.parent, File):
            return os.path.basename(self.path)
        relpath = os.path.relpath(self.path, self.environment.base_dir)
        if not relpath.startswith(".."):
            return relpath
        return os.path.abspath(self.path)

    def last_updated(self, key="st_mtime"):
        if not os.path.exists(self.path):
            return None
        return getattr(os.stat(self.path), key)


class SyncDirectory(Component):

    namevar = "path"
    source = None
    exclude = ()

    verify_opts = "-rclnv"
    sync_opts = "--inplace -lr"

    def configure(self):
        self.path = self.map(self.path)
        self.source = os.path.normpath(
            os.path.join(self.root.defdir, self.source)
        )

    @property
    def exclude_arg(self):
        if not self.exclude:
            return ""
        return " ".join("--exclude '{}'".format(x) for x in self.exclude) + " "

    def verify(self):
        if not os.path.isdir(self.path):
            raise batou.UpdateNeeded()

        stdout, stderr = self.cmd(
            "rsync {} {}{}/ {}".format(
                self.verify_opts, self.exclude_arg, self.source, self.path
            )
        )

        # In case of we see non-convergent rsync runs
        output.annotate("rsync result:", debug=True)
        output.annotate(stdout, debug=True)

        if len(stdout.strip().splitlines()) - 4 > 0:
            raise batou.UpdateNeeded()

    def update(self):
        self.cmd(
            "rsync {} {}{}/ {}".format(
                self.sync_opts, self.exclude_arg, self.source, self.path
            )
        )

    @property
    def namevar_for_breadcrumb(self):
        if isinstance(self.parent, Directory):
            return os.path.basename(self.path)
        relpath = os.path.relpath(self.path, self.environment.base_dir)
        if not relpath.startswith(".."):
            return relpath
        return os.path.abspath(self.path)


class Directory(Component):

    namevar = "path"
    leading = False
    source = None
    exclude = ()

    verify_opts = None
    sync_opts = None

    def configure(self):
        self.path = self.map(self.path)
        if self.source:
            # XXX The ordering is wrong. SyncDirectory should run *after*.
            args = {
                "source": self.source,
                "exclude": self.exclude,
            }
            if self.verify_opts:
                args["verify_opts"] = self.verify_opts
            if self.sync_opts:
                args["sync_opts"] = self.sync_opts
            self += SyncDirectory(self.path, **args)

    def verify(self):
        assert os.path.isdir(self.path)

    def update(self):
        ensure_path_nonexistent(self.path)
        if self.leading:
            os.makedirs(self.path)
        else:
            os.mkdir(self.path)

    def last_updated(self, key="st_mtime"):
        newest = 0  # epoch
        for dirpath, dirnames, filenames in os.walk(self.path):
            for filename in filenames:
                time = getattr(os.stat(os.path.join(dirpath, filename)), key)
                if time > newest:
                    newest = time
        return newest

    @property
    def namevar_for_breadcrumb(self):
        if isinstance(self.parent, File):
            return os.path.basename(self.path)
        relpath = os.path.relpath(self.path, self.environment.base_dir)
        if not relpath.startswith(".."):
            return relpath
        return os.path.abspath(self.path)


class FileComponent(Component):

    namevar = "path"
    leading = False

    def configure(self):
        self.original_path = self.path
        self.path = self.map(self.original_path)

    @property
    def namevar_for_breadcrumb(self):
        if isinstance(self.parent, File):
            return os.path.basename(self.path)
        relpath = os.path.relpath(self.path, self.environment.base_dir)
        if not relpath.startswith(".."):
            return relpath
        return os.path.abspath(self.path)


def limited_buffer(iterator, limit, lead, separator="...", logdir="/tmp"):
    limit_triggered = False
    # Fill up to limit lines into the start buffer
    start_buffer = []
    for line in iterator:
        line = line.rstrip()
        start_buffer.append(line)
        if len(start_buffer) > limit:
            break

    # Fill the remainder into the end buffer but only keep lead size.
    # This is a memory optimization: don't ever keep the whole iterator in
    # memory!
    end_buffer = []
    _, diff_log = tempfile.mkstemp(suffix=".diff", dir=logdir)
    diff_log_file = open(diff_log, "a+")
    for line in start_buffer:
        diff_log_file.write(line + "\n")
    try:
        for line in iterator:
            line = line.rstrip()
            diff_log_file.write(line + "\n")
            end_buffer.append(line)
            if len(end_buffer) > lead:
                end_buffer.pop(0)
    finally:
        diff_log_file.close()
    # If we ended up with output in the end buffer, we need to merge the
    # output.
    if end_buffer:
        start_buffer = start_buffer[:lead] + [separator] + end_buffer
        limit_triggered = True

    return start_buffer, limit_triggered, diff_log


class ManagedContentBase(FileComponent):
    """A base class that can be customized for different
    ways of structuring / managing / updating content
    of files.

    Not intended for direct use.
    """

    content = None
    source = ""
    sensitive_data = None

    # If content is given as unicode (always the case with templates)
    # then require it to be encodable. We assume UTF-8 as a sensible default
    # for most use cases and allow overrides.
    encoding = "utf-8"

    _delayed = False
    _max_diff = 200
    _max_diff_lead = 50

    _content_source_attribute = "content"

    def configure(self):
        super(ManagedContentBase, self).configure()

        self.diff_dir = os.path.join(
            self.environment.workdir_base, ".batou-diffs"
        )
        # Step 1: Determine content attribute:
        # - it might be given directly (content='...'),
        # - we might have been passed a filename (source='...'), or
        # - we might fall back using the path attribute (namevar)
        if self.source and getattr(self, self._content_source_attribute):
            raise ValueError(
                'Only one of either "{}" or "source" are allowed.'.format(
                    self._content_source_attribute
                ),
            )

        if not getattr(self, self._content_source_attribute):
            if not self.source:
                self.source = self.original_path

            if not self.source.startswith("/"):
                self.source = os.path.join(self.root.defdir, self.source)

        self._render()

    def _render(self):
        # Phase 1: acquire the source data into self.content
        if self.source:
            if os.path.exists(self.source):
                with open(
                    self.source,
                    "r" if self.encoding else "rb",
                    encoding=self.encoding,
                ) as f:
                    self.content = f.read()
            else:
                if self._delayed:
                    raise FileNotFoundError(
                        "Could not find source file {}".format(self.source)
                    )
                # We need to try rendering again later.
                self._delayed = True
                return

        # Phase 2: Decode, if we have an encoding.
        if self.content and self.encoding and not isinstance(self.content, str):
            self.content = self.content.decode(self.encoding)

        # Phase 3: We have the source content, now allow a subclass
        # to perform other operations to generate the final output.
        self.render()

        # Phase 4: If we have an encoding, encode the content (again)
        if self.encoding:
            self.content = self.content.encode(self.encoding)

    def verify(self, predicting=False):
        try:
            if self._delayed:
                self._render()
        except FileNotFoundError:
            if predicting:
                # During prediction runs we accept that delayed rending may
                # not yet work and that we will change. We might want to
                # turn this into an explicit flag so we don't implicitly
                # run into a broken deployment.
                assert False
            # If we are not predicting then this is definitely a problem.
            # Stop here.
            raise
        try:
            with open(self.path, "rb") as target:
                current = target.read()
                if current == self.content:
                    return
        except FileNotFoundError:
            current = b""
        except Exception:
            output.annotate("Unknown content - can't predict diff.")
            raise batou.UpdateNeeded()

        if not self.encoding:
            output.annotate("Not showing diff for binary data.", yellow=True)
            raise batou.UpdateNeeded()
        elif self.sensitive_data:
            output.annotate(
                "Not showing diff as it contains sensitive data.", red=True
            )
            raise batou.UpdateNeeded()

        current_text = current.decode(self.encoding, errors="replace")
        wanted_text = self.content.decode(self.encoding, errors="replace")
        current_lines = current_text.splitlines()
        wanted_lines = wanted_text.splitlines()

        contains_secrets = False
        if self.sensitive_data is None:
            words = set(
                itertools.chain(
                    *(x.split() for x in current_lines),
                    *(x.split() for x in wanted_lines),
                )
            )
            contains_secrets = bool(
                self.environment.secret_data.intersection(words)
            )

        diff = difflib.unified_diff(current_lines, wanted_lines)
        if not os.path.exists(self.diff_dir):
            os.makedirs(self.diff_dir)
        diff, diff_too_long, diff_log = limited_buffer(
            diff, self._max_diff, self._max_diff_lead, logdir=self.diff_dir
        )

        if contains_secrets:
            output.line(
                "Not showing diff as it contains sensitive data,",
                yellow=True,
            )
            output.line(f"see {diff_log} for the diff.".format(), yellow=True)
            raise batou.UpdateNeeded()

        if diff_too_long:
            output.line(
                f"More than {self._max_diff} lines of diff. Showing first "
                f"and last {self._max_diff_lead} lines.",
                yellow=True,
            )
            output.line(
                f"see {diff_log} for the full diff.".format(), yellow=True
            )

        for line in diff:
            line = line.replace("\n", "")
            if not line.strip():
                continue
            output.annotate(
                f"  {os.path.basename(self.path)} {line}",
                red=line.startswith("-"),
                green=line.startswith("+"),
            )

        raise batou.UpdateNeeded()

    def update(self):
        with open(self.path, "wb") as target:
            target.write(self.content)


class Content(ManagedContentBase):
    """Manage the content of a file - possibly using templating."""

    is_template = File.is_template
    template_context = None
    template_args = None  # dict, actually

    def render(self):
        if not self.is_template:
            return
        if self.template_args is None:
            self.template_args = dict()
        if not self.template_context:
            self.template_context = self.parent
        self.content = self.expand(
            self.content, self.template_context, args=self.template_args
        )


class JSONContent(ManagedContentBase):

    # Data to be used.
    data = None

    # Data to override the source.
    override = None

    # Cause the rendered data to be human readable. Not all parsers support
    # this so it can be turned off.
    human_readable = True

    # The final content that will be written out to the target.
    content_compact = None
    content_readable = None

    _content_source_attribute = "data"

    def render(self):
        if self.content:
            self.data = json.loads(self.content)
        if self.override:
            self.data = dict_merge(self.data, self.override)

        self.content_compact = json.dumps(
            self.data, sort_keys=True, separators=(",", ":")
        )
        self.content_readable = json.dumps(self.data, sort_keys=True, indent=4)

        if self.human_readable:
            self.content = self.content_readable
        else:
            self.content = self.content_compact


class YAMLContent(ManagedContentBase):

    # Data to be used.
    data = None

    # Data to override the source.
    override = None

    _content_source_attribute = "data"

    def render(self):
        if self.content:
            self.data = yaml.safe_load(self.content)
        if self.override:
            self.data = dict_merge(self.data, self.override)

        self.content = yaml.safe_dump(self.data)


class Owner(FileComponent):

    owner = None

    def verify(self):
        assert os.path.exists(self.path)
        if isinstance(self.owner, str):
            self.owner = pwd.getpwnam(self.owner).pw_uid
        assert os.stat(self.path).st_uid == self.owner

    def update(self):
        group = os.stat(self.path).st_gid
        os.chown(self.path, self.owner, group)


class Group(FileComponent):

    group = None

    def verify(self):
        assert os.path.exists(self.path)
        if isinstance(self.group, str):
            self.group = grp.getgrnam(self.group).gr_gid
        assert os.stat(self.path).st_gid == self.group

    def update(self):
        owner = os.stat(self.path).st_uid
        os.chown(self.path, owner, self.group)


def convert_mode(string: str) -> int:
    """Convert ls-string representation to bitmask."""
    pattern = re.compile(
        r"""^ # ensure length
        (?P<o400>r|-) # Use groups as octal values.
        (?P<o200>w|-)
        (?P<o100>x|-)
        (?P<o040>r|-)
        (?P<o020>w|-)
        (?P<o010>x|-)
        (?P<o004>r|-)
        (?P<o002>w|-)
        (?P<o001>x|-)
        $ # ensure length""",
        re.VERBOSE,
    )
    match = pattern.match(string)
    if match:
        return sum(
            int(f"0{key}", base=8)
            for key, value in match.groupdict().items()
            if value != "-"
        )
    else:
        # No match, treat it as wrong syntax
        raise SyntaxError(
            "Mode-string should be between `---------` and `rwxrwxrwx`."
        )


class Mode(FileComponent):

    mode = Attribute(default=None)

    def configure(self):
        super().configure()

        if isinstance(self.mode, str):
            try:
                self.mode = int(self.mode, 8)
            except ValueError:
                try:
                    self.mode = convert_mode(self.mode)
                except Exception as e:
                    raise batou.ConversionError.from_context(
                        self, "mode", self.mode, convert_mode, e
                    )

        elif isinstance(self.mode, int):
            pass
        else:
            raise batou.ConfigurationError.from_context(
                f"`mode` is required and `{self.mode!r}` is not a valid value.`"
            )

    def verify(self):
        try:
            self._select_stat_implementation()
        except AttributeError:
            # Happens on systems without lstat/lchmod implementation (like
            # Linux) Not sure whether ignoring it is really the right thing.
            return
        assert os.path.lexists(self.path)
        current = self._stat(self.path).st_mode
        assert stat.S_IMODE(current) == self.mode

    def update(self):
        self._chmod(self.path, self.mode)

    def _select_stat_implementation(self):
        self._stat = os.stat
        self._chmod = os.chmod
        if os.path.islink(self.path):
            self._stat = os.lstat
            self._chmod = os.lchmod


class Symlink(Component):

    namevar = "target"
    source = None

    def configure(self):
        self.target = self.map(self.target)
        self.source = self.map(self.source)

    def verify(self):
        assert os.path.islink(self.target)
        assert os.readlink(self.target) == self.source

    def update(self):
        ensure_path_nonexistent(self.target)
        os.symlink(self.source, self.target)


class Purge(Component):
    """Ensure that a set of files (given as a glob) does not exist."""

    namevar = "pattern"

    def configure(self):
        self.pattern = self.map(self.pattern)

    def verify(self):
        assert not glob.glob(self.pattern)

    def update(self):
        for filename in glob.glob(self.pattern):
            if os.path.isdir(filename):
                shutil.rmtree(filename)
            else:
                os.remove(filename)



# -*- coding: utf-8 -*-
import collections
import os.path

import jinja2
import mock
import pytest

from batou import TemplatingError
from batou.template import TemplateEngine

Server = collections.namedtuple("Server", ["name", "address"])

sample_dict = dict(
    host=mock.Mock(),
    config=dict(),
    servers=[Server("s1", "1.2.3.4"), Server("s2", "2.3.4.5")],
    hello="world",
    hello2="wörld",
)

fixture = "%s/fixture/template" % os.path.dirname(__file__)


def test_unknown_template_engine():
    with pytest.raises(NotImplementedError):
        TemplateEngine.get("foo")


def test_abstract_engine():
    engine = TemplateEngine()
    with pytest.raises(NotImplementedError):
        engine._render_template_file("asdf", {})
    with pytest.raises(NotImplementedError):
        engine.expand("asdf", {})


def test_jinja2_template_str():
    tmpl = TemplateEngine.get("jinja2")
    assert "hello world" == tmpl.expand("hello {{hello}}", sample_dict)
    assert "hello world\n\n   " == tmpl.expand(
        "hello {{hello}}\n\n   ", sample_dict
    )


def test_jinja2_template_keep_trailing_newline():
    tmpl = TemplateEngine.get("jinja2")
    assert "hello world\n" == tmpl.expand("hello world\n", {})


def test_jinja2_template_dont_add_trailing_newline():
    tmpl = TemplateEngine.get("jinja2")
    assert "hello world" == tmpl.expand("hello world", {})


def test_jinja2_template_file():
    tmpl = TemplateEngine.get("jinja2")
    filename = "{}/haproxy.cfg".format(fixture)
    result = tmpl.template(filename + ".jinja2", sample_dict)
    with open(filename) as ref:
        assert ref.read() == result


@mock.patch("batou.remote_core.Output.line")
def test_jinja2_large_template_str(output):
    tmpl = TemplateEngine.get("jinja2")
    result = tmpl.expand("hello {{hello}}" * 15000, sample_dict)
    assert result.startswith("hello world")

    log = "\n".join(c[0][0].strip() for c in output.call_args_list)

    assert (
        """\
ERROR: You are trying to render a template that is bigger than 100KiB we've \
seen that Jinja can crash at large templates and suggest you find \
alternatives for this. The affected template starts with:
hello {{hello}}hello {{hello}}hello {{hello}}hello {{hello}}hello \
{{hello}}hello {{hello}}hello {{he"""
        == log
    )


def test_jinja2_unknown_variable_should_fail():
    tmpl = TemplateEngine.get("jinja2")
    with pytest.raises(TemplatingError):
        tmpl.expand("unknown variable {{foo}}", {})


def test_jinja2_umlaut_variables():
    tmpl = TemplateEngine.get("jinja2")
    assert "hello wörld" == tmpl.expand("hello {{hello2}}", sample_dict)



import base64
import errno
import fcntl
import os
import pathlib
import pty
import subprocess
import sys
import tempfile
from typing import Dict, List, Optional

from configupdater import ConfigUpdater

from batou import AgeCallError, FileLockedError, GPGCallError
from batou._output import output

debug = False


class EncryptedFile:
    file_ending = None

    def __init__(self, path: "pathlib.Path", writeable: bool = False):
        self.path = path
        self.writeable = writeable
        self.fd = None
        self.is_new: Optional[bool] = None
        self._decrypted: Optional[bytes] = None

    @property
    def decrypted(self) -> bytes:
        if self.is_new:
            self._decrypted = b""
        if self.path.stat().st_size == 0:
            self._decrypted = b""
        if self._decrypted is None:
            self._decrypted = self.decrypt()
        if self._decrypted is None:
            raise ValueError("No decrypted data available")
        return self._decrypted

    def decrypt(self) -> bytes:
        raise NotImplementedError("decrypt() not implemented")

    @property
    def cleartext(self) -> str:
        return self.decrypted.decode("utf-8")

    @property
    def locked(self) -> bool:
        return self.fd is not None

    def write(
        self, content: bytes, recipients: List[str], reencrypt: bool = False
    ):
        if debug:
            print(
                f"EncryptedFile({self.path}).write({content}, {recipients}, {reencrypt})",
                file=sys.stderr,
            )
        self._decrypted = None
        self._write(content, recipients, reencrypt)
        self._decrypted = content

    def _write(
        self, content: bytes, recipients: List[str], reencrypt: bool = False
    ):
        raise NotImplementedError("_write() not implemented")

    def __enter__(self):
        self._lock()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._unlock()

    def _lock(self):
        if self.locked:
            raise FileLockedError.from_context(self.path)
        if not self.path.exists():
            self.is_new = True
            self.path.touch()
        self.fd = open(self.path, "r+" if self.writeable else "r")
        if debug:
            print(f"Locking `{self.path}`", file=sys.stderr)
        try:
            fcntl.lockf(
                self.fd,
                fcntl.LOCK_NB  # non-blocking
                | (
                    fcntl.LOCK_EX  # exclusive
                    if self.writeable
                    else fcntl.LOCK_SH  # shared
                ),
            )
        except BlockingIOError:
            raise FileLockedError.from_context(self.path)

    def _unlock(self):
        if debug:
            print(f"Unlocking `{self.path}`", file=sys.stderr)
        if self.fd is not None:
            self.fd.close()
            self.fd = None
        if self.is_new:
            self.path.unlink()

    @property
    def exists(self):
        return self.path.exists()

    def delete(self):
        self.path.unlink()


class NoBackingEncryptedFile(EncryptedFile):
    def __init__(self):
        super().__init__(pathlib.Path("/dev/null"))
        self.is_new = True

    def decrypt(self):
        return b""

    @property
    def locked(self):
        return True

    def _lock(self):
        pass

    def _unlock(self):
        pass


class GPGEncryptedFile(EncryptedFile):
    file_ending = ".gpg"

    def decrypt(self):
        if not self.locked:
            raise RuntimeError("File not locked")
        args = [self.gpg(), "--decrypt", str(self.path)]

        if debug:
            print(f"Running `{args}`", file=sys.stderr)

        try:
            p = subprocess.run(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise GPGCallError.from_context(
                e.cmd, e.returncode, e.stderr
            ) from e
        return p.stdout

    def _write(
        self, content: bytes, recipients: List[str], reencrypt: bool = False
    ):
        if not self.locked:
            raise RuntimeError("File not locked")
        if not self.writeable:
            raise RuntimeError("File not writeable")
        self.path.rename(str(self.path) + ".old")
        old_path = pathlib.Path(str(self.path) + ".old")
        # starting with python 3.8, pathlib.Path's rename() method
        # returns the new path, so we need to store the old path
        args = [self.gpg(), "--encrypt"]
        for recipient in recipients:
            args.extend(["-r", recipient])
        args.extend(["-o", str(self.path)])

        if debug:
            print(f"Running `{args}`", file=sys.stderr)

        try:
            subprocess.run(
                args,
                input=content,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            old_path.rename(self.path)
            raise GPGCallError.from_context(e.cmd, e.returncode, e.stderr)
        else:
            old_path.unlink()
            self.is_new = False

    _gpg = None
    GPG_BINARY_CANDIDATES = ["gpg", "gpg2"]

    @classmethod
    def gpg(cls):
        if cls._gpg is not None:
            return cls._gpg
        with tempfile.TemporaryFile() as null:
            for gpg in cls.GPG_BINARY_CANDIDATES:
                args = [gpg, "--version"]

                if debug:
                    print(f"Running `{args}`", file=sys.stderr)

                try:
                    subprocess.check_call(args, stdout=null, stderr=null)
                except (subprocess.CalledProcessError, OSError):
                    pass
                else:
                    cls._gpg = gpg
                    return cls._gpg
        raise RuntimeError(
            "Could not find gpg binary."
            " Is GPG installed? I tried looking for: {}".format(
                ", ".join("`{}`".format(x) for x in cls.GPG_BINARY_CANDIDATES)
            )
        )


def expect(fd, expected):
    """
    Expect a certain string on the given file descriptor.
    Returns a tuple of (bool, bytes) where the bool indicates whether the
    expected string was found and the actual output bytes.
    """
    try:
        actual = os.read(fd, len(expected))
    except OSError as e:
        if e.errno == errno.EIO:
            return (False, b"")
        raise
    return (actual == expected, actual)


identities = None


def get_identities():
    global identities
    if identities is None:
        identities = os.environ.get("BATOU_AGE_IDENTITIES")
        identities = (
            [x.strip() for x in identities.split(",")] if identities else []
        )
        if not identities:
            # ssh uses ~/.ssh/id_rsa,
            #  ~/.ssh/id_ecdsa, ~/.ssh/id_ecdsa_sk, ~/.ssh/id_ed25519,
            #  ~/.ssh/id_ed25519_sk and ~/.ssh/id_dsa
            # in that order.
            identities = [
                "~/.ssh/id_rsa",
                "~/.ssh/id_ecdsa",
                "~/.ssh/id_ecdsa_sk",
                "~/.ssh/id_ed25519",
                "~/.ssh/id_ed25519_sk",
                "~/.ssh/id_dsa",
            ]
        # filter on existing files
        identities = [
            os.path.expanduser(x)
            for x in identities
            if os.path.exists(os.path.expanduser(x))
        ]
        if debug:
            print(f"Found identities: {identities}", file=sys.stderr)
    return identities


known_passphrases: Dict[str, str] = {}


def get_passphrase(identity: str) -> str:
    """Prompt the user for a passphrase if necessary."""
    if identity in known_passphrases:
        return known_passphrases[identity]

    op = os.environ.get("BATOU_AGE_IDENTITY_PASSPHRASE")

    if op and not op.startswith("op://"):
        passphrase = op
    if op:
        op_process = subprocess.run(
            ["op", "read", op],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
        passphrase = op_process.stdout.decode("utf-8").strip()
    else:
        import getpass

        passphrase = getpass.getpass(
            "Enter passphrase for {}: ".format(identity)
        )

    known_passphrases[identity] = passphrase
    return passphrase


class AGEEncryptedFile(EncryptedFile):
    file_ending = ".age"

    def decrypt(self):
        if not self.locked:
            raise ValueError("File is not locked")
        identities = get_identities()
        exceptions = []
        for identity in identities:
            with tempfile.NamedTemporaryFile() as temp_file:
                args = [
                    self.age(),
                    "-d",
                    "-i",
                    str(identity),
                    "-o",
                    str(temp_file.name),
                    str(self.path),
                ]

                if debug:
                    print(f"Running `{args}`", file=sys.stderr)

                child_pid, fd = pty.fork()

                IS_CHILD = child_pid == 0

                if IS_CHILD:
                    os.execvp(args[0], args)

                assert not IS_CHILD

                matches, out = expect(
                    fd,
                    b'Enter passphrase for "'
                    + identity.encode("utf-8")
                    + b'": ',
                )

                if matches:
                    passphrase = get_passphrase(identity)
                    os.write(fd, passphrase.encode("utf-8") + b"\n")
                    matches, out = expect(fd, b"\r\r\n")
                    if not matches:
                        exceptions.append(
                            Exception(
                                'Unexpected output from age, expected "\\r\\r\\n": {}'.format(
                                    out
                                )
                            )
                        )
                        continue
                    # also assert, that output is empty from now on
                    buffer = b""
                    while True:
                        try:
                            chunk = os.read(fd, 1024)
                        except OSError as err:  # noqa
                            if err.errno == errno.EIO:
                                # work arond suspected pty "feature", where
                                # reading from the file descriptor
                                # when there is no data raises instead of
                                # returning an empty string, see
                                # https://bugs.python.org/issue5380
                                chunk = None
                        if not chunk:
                            break
                        buffer += chunk
                    if buffer:
                        magic_bytes = b"\x1b[F\x1b[K"
                        if buffer.startswith(magic_bytes):
                            buffer = buffer[len(magic_bytes) :]
                    if buffer:
                        exceptions.append(
                            Exception(
                                "Unexpected output from age: {}".format(buffer)
                            )
                        )
                        continue

                # Wait for the child to exit
                pid, exitcode = os.waitpid(child_pid, 0)

                if exitcode != 0 or pid != child_pid:
                    exceptions.append(
                        AgeCallError.from_context(args, exitcode, out)
                    )
                    continue

                temp_file.seek(0)
                result = temp_file.read()

                if result:
                    return result
        for e in exceptions:
            print(e)
        raise Exception(
            f"Could not decrypt {self.path} with any of the identities {identities}"
        )

    def _write(
        self, content: bytes, recipients: List[str], reencrypt: bool = False
    ):
        if not self.locked:
            raise ValueError("File is not locked")
        if not self.writeable:
            raise ValueError("File is not writeable")
        args = [self.age(), "-e"]
        for recipient in recipients:
            args.extend(["-r", recipient])
        args.extend(["-o", str(self.path)])

        if debug:
            print(f"Running `{args}`", file=sys.stderr)

        try:
            p = subprocess.run(
                args,
                input=content,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise AgeCallError.from_context(e.cmd, e.returncode, e.stderr)
        self.is_new = False

    _age = None
    AGE_BINARY_CANDIDATES = ["age", "rage"]

    @classmethod
    def age(cls):
        if cls._age is not None:
            return cls._age
        with tempfile.TemporaryFile() as null:
            for age in cls.AGE_BINARY_CANDIDATES:
                args = [age, "--version"]
                if debug:
                    print(f"Running `{args}`", file=sys.stderr)
                try:
                    subprocess.check_call(args, stdout=null, stderr=null)
                except (subprocess.CalledProcessError, OSError):
                    pass
                else:
                    cls._age = age
                    return cls._age
        raise RuntimeError(
            "Could not find age binary."
            " Is age installed? I tried looking for: {}".format(
                ", ".join("`{}`".format(x) for x in cls.AGE_BINARY_CANDIDATES)
            )
        )


class DiffableAGEEncryptedFile(EncryptedFile):
    file_ending = ".age-diffable"

    def __init__(self, path: "pathlib.Path", writeable: bool = False):
        super().__init__(path, writeable)
        self._decrypted_content = None
        self._encrypted_content = None

    def decrypt_age_string(self, content: str) -> str:
        # base64 -> tmpfile -> AGEEncryptedFile -> decrypt -> read
        with tempfile.NamedTemporaryFile() as temp_file:
            temp_file.write(base64.b64decode(content))
            temp_file.flush()
            with AGEEncryptedFile(pathlib.Path(temp_file.name)) as ef:
                return ef.cleartext

    def encrypt_age_string(self, content: str, recipients: List[str]) -> str:
        # tmpfile -> AGEEncryptedFile -> write plaintext -> read ciphertext -> base64
        with tempfile.NamedTemporaryFile() as temp_file:
            with AGEEncryptedFile(pathlib.Path(temp_file.name), True) as ef:
                ef.write(content.encode("utf-8"), recipients)
            with open(temp_file.name, "rb") as f:
                return base64.b64encode(f.read()).decode("utf-8")

    def decrypt(self):
        # read the entire file, parse as ConfigUpdater

        if not self.locked:
            raise ValueError("File is not locked")

        config_encrypted = ConfigUpdater().read(self.path)
        config = ConfigUpdater().read(self.path)

        # for each section
        for section in config.sections():
            # if section is called batou, skip
            if section == "batou":
                continue
            # for each option
            for option in config[section]:
                # decrypt the value
                decrypted = self.decrypt_age_string(
                    config[section][option].value
                )
                if "\n" in decrypted:
                    # multiline: accounts for indents
                    config[section][option].set_values(
                        decrypted.split("\n"),
                        prepend_newline=False,
                    )
                else:
                    config[section][option].value = decrypted

        # cache the decrypted content
        self._decrypted_content = config
        self._encrypted_content = config_encrypted

        # return the decrypted content as bytes
        return str(config).encode("utf-8")

    def _write(
        self, content: bytes, recipients: List[str], reencrypt: bool = False
    ):
        # parse the content as ConfigUpdater
        config = ConfigUpdater()
        config.read_string(content.decode("utf-8"))

        # for each section
        for section in config.sections():
            # if section is called batou, skip
            if section == "batou":
                continue
            # for each option
            for option in config[section]:
                new_value = config[section][option].value
                assert new_value is not None

                try:
                    old_value = self._decrypted_content[section][option].value
                except Exception:
                    old_value = None

                value_has_changed = new_value != old_value

                if reencrypt or value_has_changed:
                    new_encrypted_value = self.encrypt_age_string(
                        new_value, recipients
                    )
                else:
                    new_encrypted_value = self._encrypted_content[section][
                        option
                    ].value

                config[section][option].value = new_encrypted_value

        # write the config to the file
        with open(self.path, "w") as f:
            f.write(str(config))

        self.is_new = False


all_encrypted_file_types = [
    NoBackingEncryptedFile,
    GPGEncryptedFile,
    AGEEncryptedFile,
    DiffableAGEEncryptedFile,
]


def get_encrypted_file(
    path: "pathlib.Path", writeable: bool = False
) -> EncryptedFile:
    """Return the appropriate EncryptedFile object for the given path."""
    for ef in all_encrypted_file_types:
        if ef.file_ending and path.name.endswith(ef.file_ending):
            return ef(path, writeable)
    raise ValueError(f"Unknown encrypted file type for {path}")



# This should be only one line. If it must be multi-line, indent the second
# line onwards to keep the PKG-INFO file format intact.
"""A utility for automating multi-host, multi-environment software builds \
and deployments.
"""

from setuptools import find_packages, setup

version = open("src/batou/version.txt").read().strip()

setup(
    name="batou",
    version=version,
    install_requires=[
        "ConfigUpdater>=3.2",
        "Jinja2>=3.1.4",
        "requests",
        # ConfigUpdater does not manage its minimum requirements correctly.
        "setuptools>=66.1.0",
        "execnet>=1.8.1",
        "importlib_metadata",
        "importlib_resources",
        "py>=1.11.0",
        "pyyaml",
        "remote-pdb",
    ],
    extras_require={
        "test": [
            "mock",
            "pytest",
            "pytest-coverage",
            "pytest-instafail",
            "pytest-timeout",
        ]
    },
    entry_points="""
        [console_scripts]
            batou = batou.main:main
        [zc.buildout]
            requirements = batou.buildout:Requirements
        [batou.provisioners]
            fc-nixos-dev-container = batou.provision:FCDevContainer
            fc-nixos-dev-vm = batou.provision:FCDevVM
    """,
    author="Christian Theune",
    author_email="ct@flyingcircus.io",
    license="BSD (2-clause)",
    url="https://batou.readthedocs.io/en/latest/",
    keywords="deployment",
    classifiers="""\
License :: OSI Approved :: BSD License
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3 :: Only
"""[
        :-1
    ].split(
        "\n"
    ),
    description=__doc__.strip(),
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    test_suite="batou.tests",
    python_requires=">=3.6",
)



import os.path

from mock import Mock

from batou.vfs import Developer, Map


def test_developer_mapping(tmpdir):
    environment = Mock()
    environment.workdir_base = str(tmpdir)
    config = Mock()
    mapping = Developer(environment, config)

    assert mapping.map("/etc") == str(tmpdir / "_" / "etc")
    assert os.path.exists(str(tmpdir / "_"))
    assert not os.path.exists(str(tmpdir / "_" / "etc"))

    assert mapping.map("/var") == str(tmpdir / "_" / "var")
    assert not os.path.exists(str(tmpdir / "_" / "var"))

    assert mapping.map(str(tmpdir / "foo")) == str(tmpdir / "foo")
    assert mapping.map("asdf") == "asdf"


def test_arbitrary_mapping():
    environment = Mock()
    environment.workdir_base = "workdir"
    config = {
        "/a": "/a1",
        # longer paths are preferred matches
        "/a/b": "/a2",
        # non-absolute paths are ignored
        "etc": "none",
    }
    mapping = Map(environment, config)

    assert mapping.map("/etc") == "/etc"
    assert mapping.map("/a") == "/a1"
    assert mapping.map("/a/b/c") == "/a2/c"
    assert mapping.map("/a/c") == "/a1/c"
    assert mapping.map("workdir/foo") == "workdir/foo"



import mock

from batou import NonConvergingWorkingSet, UnsatisfiedResources
from batou.environment import Environment
from batou.resources import Resources


def test_regression_reset_works_if_provided_but_not_required():
    resources = Resources()
    root = mock.Mock()
    resources.provide(root, mock.sentinel.key, "asdf")
    resources.reset_component_resources(root)


def test_reset_marks_depending_components_as_dirty():
    resources = Resources()
    root1 = mock.Mock()
    resources.provide(root1, mock.sentinel.key, "asdf")

    root2 = mock.Mock()
    resources.require(root2, mock.sentinel.key)

    root3 = mock.Mock()
    resources.require(root3, mock.sentinel.key, dirty=True)
    assert resources.dirty_dependencies == set()
    resources.reset_component_resources(root1)
    assert resources.dirty_dependencies == set([root2])


def test_mentions_missing_requirement_with_host_requirement(sample_service):
    e = Environment("test-resources-host")
    e.load()
    errors = e.configure()
    assert len(errors) == 2
    assert isinstance(errors[0], UnsatisfiedResources)
    assert isinstance(errors[1], NonConvergingWorkingSet)
    assert errors[0].unsatisfied_resources == [
        ("key", "host1", []),
        ("unrelated", None, ["component1"]),
    ]



# -*- coding: utf-8 -*-

import getpass
import grp
import json
import os
import pwd
from stat import S_IMODE

import pytest
import yaml
from mock import Mock, patch

import batou
from batou.lib.file import (
    BinaryFile,
    Content,
    Directory,
    File,
    FileComponent,
    JSONContent,
    Mode,
    Presence,
    Purge,
    Symlink,
    SyncDirectory,
    YAMLContent,
    ensure_path_nonexistent,
)
from batou.tests.ellipsis import Ellipsis


def test_ensure_path_nonexistent_removes_normal_file(tmpdir):
    open("asdf", "w").close()
    assert os.path.exists("asdf")
    ensure_path_nonexistent("asdf")
    assert not os.path.exists("asdf")


def test_ensure_path_nonexistent_removes_normal_symlink(tmpdir):
    os.chdir(str(tmpdir))
    open("target", "w").close()
    os.symlink("target", "link")
    assert os.path.exists("link")
    assert os.path.exists("target")

    ensure_path_nonexistent("link")

    assert not os.path.exists("link")
    assert os.path.exists("target")


def test_ensure_path_nonexistent_removes_broken_symlink(tmpdir):
    os.chdir(str(tmpdir))
    os.symlink("target", "link")
    assert os.path.islink("link")
    assert not os.path.exists("target")
    ensure_path_nonexistent("link")
    assert not os.path.exists("link")
    assert not os.path.exists("target")


def test_ensure_path_nonexistent_removes_directory(tmpdir):
    os.chdir(str(tmpdir))
    os.mkdir("dir")
    assert os.path.exists("dir")
    ensure_path_nonexistent("dir")
    assert not os.path.exists("dir")


def test_ensure_path_does_not_fail_on_nonexisting_path():
    assert not os.path.exists("missing")
    ensure_path_nonexistent("missing")
    assert not os.path.exists("missing")


def test_presence_creates_nonexisting_file(root):
    p = Presence("path")
    root.component += p
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == ""


def test_presence_leaves_existing_file_with_content_intact(root):
    p = Presence("path")
    root.component += p
    with open(p.path, "w") as f:
        f.write("Hello there!")
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == "Hello there!"


def test_presence_creates_directories_if_configured(root):
    p = Presence("directory/file", leading=True)
    root.component += p
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == ""


def test_presence_doesnt_create_directories_by_default(root):
    root.component += Presence("directory/file")
    with pytest.raises(IOError):
        root.component.deploy()


def test_presence_removes_conflicting_symlinks(root):
    p = Presence("link")
    root.component += p
    os.symlink("target", p.path)
    assert os.path.islink(p.path)
    with pytest.raises(IOError):
        open(p.path)
    root.component.deploy()
    assert not os.path.islink(p.path)
    with open(p.path) as f:
        assert f.read() == ""


def test_presence_removes_conflicting_directories(root):
    p = Presence("dir")
    root.component += p
    os.mkdir(p.path)
    root.component.deploy()
    assert not os.path.isdir(p.path)
    with open(p.path) as f:
        assert f.read() == ""


def test_directory_creates_directory(root):
    path = "dir"
    d = Directory(path)
    root.component += d
    assert not os.path.isdir(d.path)
    root.component.deploy()
    assert os.path.isdir(d.path)


def test_directory_creates_leading_directories_if_configured(root):
    path = "directory/path"
    d = Directory(path, leading=True)
    root.component += d
    root.component.deploy()
    assert os.path.isdir(d.path)


def test_directory_doesnt_create_leading_directories_by_default(root):
    path = "directory/path"
    root.component += Directory(path)
    with pytest.raises(OSError):
        root.component.deploy()


def test_filecomponent_baseclass_carries_path(root):
    path = "path"
    p = FileComponent(path)
    root.component += p
    assert p.path.endswith(path)
    assert p.original_path == path


def test_content_passed_by_string(root):
    path = "path"
    p = Content(path, content="asdf")
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == "asdf"


def test_content_passed_by_string_template(root):
    path = "path"
    root.component.foobar = "asdf"
    p = Content(path, content="{{component.foobar}}")
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == "asdf"


def test_content_with_unicode_requires_encoding(root):
    path = "path"
    root.component.foobar = "äsdf"
    p = File(path, content="örks {{component.foobar}}", encoding="ascii")

    with pytest.raises(UnicodeEncodeError):
        root.component |= p

    p = File(path, content="örks {{component.foobar}}", encoding="utf-8")
    root.component += p
    root.component.deploy()
    with open(p.path, encoding=p.encoding) as f:
        result = f.read()
        # XXX pytest reporting breaks if this fails. :(
        assert result == "örks äsdf"


def test_content_passed_by_string_notemplate(root):
    path = "path"
    root.component.foobar = "asdf"
    p = Content(path, content="{{component.foobar}}", is_template=False)
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == "{{component.foobar}}"


def test_content_passed_by_file(root):
    source = "source"
    with open(source, "w") as f:
        f.write("content from source file")
    path = "path"
    p = Content(path, source=source)
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == "content from source file"


def test_content_passed_by_file_handles_encoding(root):
    source = "source"
    with open(source, "w", encoding="latin-1") as f:
        f.write("cöntent from source file")
    path = "path"
    p = Content(path, source=source, encoding="latin-1")
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path, encoding="latin-1") as f:
        assert f.read() == "cöntent from source file"


def test_content_passed_by_file_handles_encoding_on_verify(root):
    source = "source"
    path = "path"
    p = Content(path, source=source, encoding="latin-1")
    root.component += p
    with open(source, "w", encoding="latin-1") as f:
        f.write("cöntent from source file")
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path, encoding="latin-1") as f:
        assert f.read() == "cöntent from source file"


def test_content_passed_by_file_defaults_to_utf8(root):
    source = "source"
    with open(source, "w", encoding="utf-8") as f:
        f.write("cöntent from source file")
    path = "path"
    p = Content(path, source=source)
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path, encoding="utf-8") as f:
        assert f.read() == "cöntent from source file"


def test_content_passed_by_file_template(root):
    source = "source"
    with open(source, "w") as f:
        f.write("{{component.foobar}}")
    path = "path"
    root.component.foobar = "asdf"
    p = Content(path, source=source, is_template=True)
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == "asdf"


def test_content_passed_by_file_template_handles_encoding(root):
    source = "source"
    with open(source, "w", encoding="latin-1") as f:
        f.write("cöntent from source file {{component.foo}}")
    path = "path"
    p = Content(path, source=source, encoding="latin-1")
    root.component.foo = "foo"
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path, encoding="latin-1") as f:
        assert f.read() == "cöntent from source file foo"


def test_content_passed_by_file_template_defaults_to_utf8(root):
    source = "source"
    with open(source, "w", encoding="utf-8") as f:
        f.write("cöntent from source file {{component.foo}}")
    path = "path"
    p = Content(path, source=source)
    root.component.foo = "foo"
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path, encoding="utf-8") as f:
        assert f.read() == "cöntent from source file foo"


def test_content_passed_by_file_no_template_is_binary(root):
    # This is a regression test for #14944 where UTF 8 in a
    # non-template file caused an accidental implicit encoding to ASCII
    source = "source"
    with open(source, "wb") as f:
        f.write(b"\x89PNG\r\n\x1a\n")
    path = "path"
    p = Content(path, source=source, is_template=False, encoding=None)
    root.component.foo = "foo"
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path, "rb") as f:
        assert f.read() == (b"\x89PNG\r\n\x1a\n")


def test_binary_file_component(root):
    # This is a regression test for #14944 where UTF 8 in a
    # non-template file caused an accidental implicit encoding to ASCII
    source = "source"
    with open(source, "wb") as f:
        f.write(b"\x89PNG\r\n\x1a\n")
    path = "path"
    p = BinaryFile(path, source=source)
    root.component.foo = "foo"
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path, "rb") as f:
        assert f.read() == (b"\x89PNG\r\n\x1a\n")


def test_content_from_file_as_template_guessed(root):
    path = "path"
    with open(path, "w") as f:
        f.write("content from source file {{component.foo}}")
    p = File(path)
    root.component.foo = "bar"
    root.component += p
    assert p.source == root.defdir + "/path"
    root.component.deploy()
    with open(p.path) as f:
        # Actually, as source and target are the same: nothing happened.
        assert f.read() == "content from source file bar"


def test_content_source_unclear(root):
    path = "path"
    p = File(path)
    with pytest.raises(ValueError) as e:
        root.component += p
    assert str(e.value) == (
        "Missing implicit template file {}/path. "
        "Or did you want to create an empty file? "
        "Then use File('path', content='').".format(root.defdir)
    )


def test_content_passed_by_file_using_path_as_default(root):
    path = "path"
    with open(path, "w") as f:
        f.write("content from source file")
    p = Content(path)
    root.component += p
    with open(p.path, "w") as f:
        # Ensure file exists
        pass
    assert p.source == root.defdir + "/path"
    root.component.deploy()
    with open(p.path) as f:
        # Actually, as source and target are the same: nothing happened.
        assert f.read() == "content from source file"


def test_content_template_with_explicit_context(root):
    path = "path"
    context = Mock()
    context.foobar = "asdf"
    p = Content(
        path,
        content="{{component.foobar}}",
        is_template=True,
        template_context=context,
    )
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == "asdf"


def test_content_relative_source_path_computed_wrt_definition_dir(root):
    path = "path"
    source = "source"
    with open(source, "w") as f:
        f.write("asdf")
    p = Content(path, source=source)
    root.component += p
    with open(p.path, "w") as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == "asdf"


def test_content_only_required_changes_touch_file(root):
    path = "path"
    p = Content(path, content="asdf")
    root.component += p
    # Lets start with an existing file that has the wrong content:
    with open(p.path, "w") as f:
        f.write("bsdf")
    os.utime(p.path, (0, 0))
    # Modify mtime so we can prove it's been changed.
    root.component.deploy()
    stat = os.stat(p.path)
    assert stat.st_mtime != 0

    os.utime(p.path, (0, 0))
    stat = os.stat(p.path)
    assert stat.st_mtime == 0

    root.component.deploy()
    stat = os.stat(p.path)
    assert stat.st_mtime == 0


def test_content_does_not_allow_both_content_and_source(root):
    path = "path"
    with pytest.raises(ValueError):
        root.component += Content(path, content="asdf", source="bsdf")


def test_content_large_diff_logged(output, root):
    path = "path"
    p = Content(path, content="\n".join(["asdf"] * 21))
    p._max_diff = 20
    p._max_diff_lead = 5
    root.component += p
    with open(p.path, "w") as f:
        f.write("\n".join(["bsdf"] * 21))
    root.component.deploy()
    log = os.listdir(p.diff_dir)[0]
    with open(os.path.join(p.diff_dir, log)) as f:
        assert (
            f.read()
            == """\
---
+++
@@ -1,21 +1,21 @@
"""
            + "\n".join(["-bsdf"] * 21)
            + "\n"
            + "\n".join(["+asdf"] * 21)
            + "\n"
        )

    assert output.backend.output == Ellipsis(
        """\
localhost > MyComponent > Content('work/mycomponent/path')
More than 20 lines of diff. Showing first and last 5 lines.
see ... for the full diff.
  path ---
  path +++
  path @@ -1,21 +1,21 @@
  path -bsdf
  path -bsdf
  path ...
  path +asdf
  path +asdf
  path +asdf
  path +asdf
  path +asdf
"""
    )


def test_json_content_data_given(root):
    p = JSONContent("target.json", data={"asdf": 1})
    root.component += p
    root.component.deploy()
    assert (
        p.content
        == b"""\
{
    "asdf": 1
}"""
    )

    with open(p.path, "rb") as f:
        assert f.read() == p.content


def test_json_diff(output, root):
    p = JSONContent("target.json", data={"asdf": 1, "bsdf": 2})
    root.component += p

    with open(p.path, "w") as f:
        assert f.write(json.dumps({"bsdf": 2}, sort_keys=True, indent=4))

    p.deploy()

    assert output.backend.output == Ellipsis(
        """\
localhost > MyComponent > JSONContent('work/mycomponent/target.json')
  target.json ---
  target.json +++
  target.json @@ -1,3 +1,4 @@
  target.json  {
  target.json +    "asdf": 1,
  target.json      "bsdf": 2
  target.json  }
"""
    )


def test_json_diff_not_for_sensitive(output, root):
    p = JSONContent(
        "target.json", data={"asdf": 1, "bsdf": 2}, sensitive_data=True
    )
    root.component += p

    with open(p.path, "w") as f:
        assert f.write(json.dumps({"bsdf": 2}, sort_keys=True, indent=4))

    p.deploy()

    assert output.backend.output == Ellipsis(
        """\
localhost > MyComponent > JSONContent('work/mycomponent/target.json')
Not showing diff as it contains sensitive data.
"""
    )


def test_json_content_data_given_compact(root):
    p = JSONContent("target.json", data={"asdf": 1}, human_readable=False)
    root.component += p
    root.component.deploy()
    assert p.content == b"""{"asdf":1}"""

    with open(p.path, "rb") as f:
        assert f.read() == p.content


def test_json_content_source_given(root):
    with open(root.defdir + "/source.json", "w", encoding="utf-8") as f:
        f.write(json.dumps([1, 2, 3, 4]))

    p = JSONContent("target.json", source="source.json")
    root.component += p
    root.component.deploy()
    assert (
        p.content
        == b"""\
[
    1,
    2,
    3,
    4
]"""
    )

    with open(p.path, "rb") as f:
        assert f.read() == p.content


def test_json_content_delayed_source_given(root):
    p = JSONContent("target.json", source="source.json")
    root.component += p

    with open(root.defdir + "/source.json", "w", encoding="utf-8") as f:
        f.write(json.dumps([1, 2, 3, 4]))

    root.component.deploy()
    assert (
        p.content
        == b"""\
[
    1,
    2,
    3,
    4
]"""
    )

    with open(p.path, "rb") as f:
        assert f.read() == p.content


def test_json_content_delayed_source_causes_predicting_verify_to_raise(root):
    p = JSONContent("target.json", source="source.json")
    root.component += p
    with pytest.raises(AssertionError):
        p.verify(predicting=True)


def test_json_content_source_with_override(root):
    with open(root.defdir + "/source.json", "w", encoding="utf-8") as f:
        f.write(
            json.dumps(
                {"database": {"address": "localhost", "password": "topsecret"}}
            )
        )

    p = JSONContent(
        "target.json",
        source="source.json",
        override={"database": {"password": "realpassword"}},
    )

    root.component += p
    root.component.deploy()

    # fmt: off
    assert (p.content == b"""\
{
    "database": {
        "address": "localhost",
        "password": "realpassword"
    }
}\
""")
    # fmt: on

    with open(p.path, "rb") as f:
        assert f.read() == p.content


def test_json_content_source_missing(root):
    p = JSONContent("target.json", source="source.json")
    root.component += p
    with pytest.raises(FileNotFoundError):
        root.component.deploy()


def test_json_content_either_source_or_data(root):
    p = JSONContent("target.json", data={"asdf": 1}, source="asdf")
    with pytest.raises(ValueError):
        root.component += p


def test_json_content_implicit_source(root):
    p = JSONContent("target.json")
    root.component += p
    assert p.source.endswith("/target.json")


def test_json_content_explicit_source(root):
    p = JSONContent("target.json", source="thesource.json")
    root.component += p
    assert p.source.endswith("/thesource.json")


def test_json_content_explicit_absolute_source(root):
    p = JSONContent("target.json", source="/tmp/thesource.json")
    root.component += p
    assert p.source == "/tmp/thesource.json"


def test_yaml_content_data_given(root):
    p = YAMLContent("target.yaml", data={"asdf": 1})
    root.component += p
    root.component.deploy()
    assert (
        p.content
        == b"""\
asdf: 1
"""
    )

    with open(p.path, "rb") as f:
        assert f.read() == p.content


def test_yaml_diff(output, root):
    p = YAMLContent("target.yaml", data={"asdf": 1, "bsdf": 2})
    root.component += p

    with open(p.path, "w") as f:
        assert f.write(json.dumps({"bsdf": 2}, sort_keys=True, indent=4))

    p.deploy()

    # fmt: off
    assert output.backend.output == Ellipsis("""\
localhost > MyComponent > YAMLContent(\'work/mycomponent/target.yaml\')
  target.yaml ---
  target.yaml +++
  target.yaml @@ -1,3 +1,2 @@
  target.yaml -{
  target.yaml -    "bsdf": 2
  target.yaml -}
  target.yaml +asdf: 1
  target.yaml +bsdf: 2
""")
    # fmt: on


def test_yaml_diff_not_for_sensitive(output, root):
    p = YAMLContent(
        "target.yaml", data={"asdf": 1, "bsdf": 2}, sensitive_data=True
    )
    root.component += p

    with open(p.path, "w") as f:
        assert f.write(json.dumps({"bsdf": 2}, sort_keys=True, indent=4))

    p.deploy()

    assert output.backend.output == Ellipsis(
        """\
localhost > MyComponent > YAMLContent('work/mycomponent/target.yaml')
Not showing diff as it contains sensitive data.
"""
    )


def test_yaml_content_source_given(root):
    with open(root.defdir + "/source.json", "w", encoding="utf-8") as f:
        f.write(json.dumps([1, 2, 3, 4]))

    p = YAMLContent("target.json", source="source.json")
    root.component += p
    root.component.deploy()
    # fmt: off
    assert p.content == b"""\
- 1
- 2
- 3
- 4
"""
    # fmt: on

    with open(p.path, "rb") as f:
        assert f.read() == p.content


def test_yaml_content_delayed_source_given(root):
    p = YAMLContent("target.yaml", source="source.yaml")
    root.component += p

    with open(root.defdir + "/source.yaml", "w", encoding="utf-8") as f:
        f.write(yaml.safe_dump([1, 2, 3, 4]))

    root.component.deploy()
    # fmt: off
    assert p.content == b"""\
- 1
- 2
- 3
- 4
"""
    # fmt: on

    with open(p.path, "rb") as f:
        assert f.read() == p.content


def test_yaml_content_delayed_source_causes_predicting_verify_to_raise(root):
    p = YAMLContent("target.json", source="source.json")
    root.component += p
    with pytest.raises(AssertionError):
        p.verify(predicting=True)


def test_yaml_content_source_with_override(root):
    with open(root.defdir + "/source.yaml", "w", encoding="utf-8") as f:
        f.write(
            yaml.safe_dump(
                {"database": {"address": "localhost", "password": "topsecret"}}
            )
        )

    p = YAMLContent(
        "target.yaml",
        source="source.yaml",
        override={"database": {"password": "realpassword"}},
    )

    root.component += p
    root.component.deploy()

    # fmt: off
    assert (p.content == b"""\
database:
  address: localhost
  password: realpassword
""")
    # fmt: on

    with open(p.path, "rb") as f:
        assert f.read() == p.content


def test_yaml_content_source_missing(root):
    p = YAMLContent("target.json", source="source.json")
    root.component += p
    with pytest.raises(FileNotFoundError):
        root.component.deploy()


def test_yaml_content_either_source_or_data(root):
    p = YAMLContent("target.json", data={"asdf": 1}, source="asdf")
    with pytest.raises(ValueError):
        root.component += p


def test_yaml_content_implicit_source(root):
    p = YAMLContent("target.json")
    root.component += p
    assert p.source.endswith("/target.json")


def test_yaml_content_explicit_source(root):
    p = YAMLContent("target.json", source="thesource.json")
    root.component += p
    assert p.source.endswith("/thesource.json")


def test_yaml_content_explicit_absolute_source(root):
    p = YAMLContent("target.json", source="/tmp/thesource.json")
    root.component += p
    assert p.source == "/tmp/thesource.json"


def test_mode_verifies_for_nonexistent_file(root):
    mode = Mode("asdf", mode=0o000)
    with pytest.raises(AssertionError):
        mode.verify()


@pytest.mark.parametrize(
    "input,expected",
    [
        (0o777, 0o777),
    ],
)
def test_mode_ensures_mode_for_files(root, input, expected):
    path = "path"
    open("work/mycomponent/" + path, "w").close()
    mode = Mode(path, mode=0o000)
    root.component += mode
    root.component.deploy()
    assert S_IMODE(os.stat(mode.path).st_mode) == 0o000

    mode.mode = 0o777
    root.component.deploy()
    assert S_IMODE(os.stat(mode.path).st_mode) == 0o777
    assert mode.changed

    root.component.deploy()
    assert not mode.changed


def test_mode_converts_to_numeric(root):
    path = "path"
    open("work/mycomponent/" + path, "w").close()

    with pytest.raises(batou.ConfigurationError) as e:
        mode = Mode(path)
        root.component += mode
    assert (
        str(e.value) == "`mode` is required and `None` is not a valid value.`"
    )

    mode = Mode(path, mode="rwx------")
    root.component += mode
    assert mode.mode == 0o700

    mode = Mode(path, mode="500")
    root.component += mode
    assert mode.mode == 0o500


def test_mode_ensures_mode_for_directories(root):
    path = "path"
    os.makedirs("work/mycomponent/path")
    mode = Mode(path, mode=0o000)
    root.component += mode
    root.component.deploy()
    assert S_IMODE(os.stat(mode.path).st_mode) == 0o000

    mode.mode = 0o777
    root.component.deploy()
    assert S_IMODE(os.stat(mode.path).st_mode) == 0o777
    assert mode.changed

    root.component.deploy()
    assert not mode.changed


@pytest.mark.skipif(not hasattr(os, "lchmod"), reason="requires lchmod")
def test_mode_ensures_mode_for_symlinks(root):
    # This test is only relevant on platforms that support managing the mode of
    # symlinks.
    link_to = "link_to"
    open(link_to, "w").close()
    os.symlink(link_to, "work/mycomponent/path")
    mode = Mode("path", mode=0o000)
    root.component += mode
    root.component.deploy()
    assert S_IMODE(os.lstat("work/mycomponent/path").st_mode) == 0o000

    mode.mode = 0o777
    root.component.deploy()
    assert S_IMODE(os.lstat("work/mycomponent/path").st_mode) == 0o777
    assert mode.changed

    root.component.deploy()
    assert not mode.changed


@pytest.mark.skipif(hasattr(os, "lchmod"), reason="requires no lchmod")
def test_mode_does_not_break_on_platforms_without_lchmod(root):
    # This test is only relevant on platforms without lchmod. We basically
    # ensure that deploying the component doesn't break but it's a noop
    # anyway.
    path = "path"
    link_to = "link_to"
    open(link_to, "w").close()
    mode = Mode(path, mode=0o000)
    root.component += mode
    os.symlink(link_to, mode.path)
    root.component.deploy()


def test_symlink_creates_new_link(root):
    link = "path"
    link_to = "link_to"
    symlink = Symlink(link, source=link_to)
    root.component += symlink
    root.component.deploy()
    assert os.readlink("work/mycomponent/" + link) == symlink.source


def test_symlink_updates_existing_link(root):
    link = "path"
    link_to = "link_to"
    # Create an initial link
    symlink = Symlink(link, source=link_to)
    root.component += symlink
    root.component.deploy()
    # Update link with other target
    link_to2 = "link_to2"
    symlink.source = link_to2
    root.component.deploy()
    assert os.readlink("work/mycomponent/" + link) == link_to2


def test_file_creates_subcomponent_for_presence(root):
    path = "path"
    file = File(path, content="")
    assert file.ensure == "file"
    root.component += file
    assert isinstance(file.sub_components[0], Presence)


def test_file_creates_subcomponent_for_directory(root):
    file = File("dir", ensure="directory")
    root.component += file
    root.component.deploy()
    assert isinstance(file.sub_components[0], Directory)


def test_file_creates_subcomponent_for_symlink(root):
    file = File("link", ensure="symlink", link_to="target")
    root.component += file
    root.component.deploy()
    assert isinstance(file.sub_components[0], Symlink)


def test_file_prohibits_unknown_ensure_parameter(root):
    with pytest.raises(ValueError):
        root.component += File("file", ensure="pipe")


@pytest.mark.slow
def test_directory_copies_all_files(root):
    os.mkdir("source")
    open("source/one", "w").close()
    open("source/two", "w").close()
    root.component += Directory("target", source="source")
    root.component.deploy()
    assert sorted(os.listdir("work/mycomponent/target")) == ["one", "two"]


@pytest.mark.slow
def test_directory_file_copies_all_files(root):
    os.mkdir("source")
    open("source/one", "w").close()
    open("source/two", "w").close()
    root.component += File("target", ensure="directory", source="source")
    root.component.deploy()
    assert sorted(os.listdir("work/mycomponent/target")) == ["one", "two"]


@pytest.mark.slow
def test_directory_last_updated_reflects_file_changes(root):
    os.mkdir("source")
    open("source/one", "w").close()
    open("source/two", "w").close()
    d = Directory("target", source="source")
    root.component += d
    root.component.deploy()
    assert (
        d.last_updated()
        == os.stat(os.path.join(root.workdir, "target", "two")).st_mtime
    )


@pytest.mark.slow
def test_directory_does_not_copy_excluded_files(root):
    os.mkdir("source")
    open("source/one", "w").close()
    open("source/two", "w").close()
    p = Directory("target", source="source", exclude=("two",))
    root.component += p
    root.component.deploy()
    assert len(os.listdir("work/mycomponent/target")) == 1


@patch("os.chown")
def test_owner_lazy(chown, root):
    with open("asdf", "w"):
        pass
    file = File("asdf", owner=getpass.getuser())
    root.component += file
    root.component.deploy()
    assert not os.chown.called


@patch("os.chown")
@patch("os.stat")
def test_owner_calls_chown(chown, stat, root):
    os.stat.return_value = Mock()
    os.stat.return_value.st_uid = 0
    os.stat.return_value.st_mode = 0
    file = File("asdf", owner=getpass.getuser(), content="")
    root.component += file
    root.component.deploy()
    assert os.chown.called


def test_owner_is_configurable_when_user_doesnt_exist_yet(root):
    file = File("asdf", owner="foobar", content="")
    # This is a regression test against #12911 and ensures that we can
    # configure a file component's owner even if the owner doesn't exist yet.
    root.component += file


def current_group():
    gid = pwd.getpwnam(getpass.getuser()).pw_gid
    for group in grp.getgrall():
        if group.gr_gid == gid:
            return group.gr_name


@patch("os.chown")
def test_group_lazy(chown, root):
    with open("asdf", "w"):
        pass
    file = File("asdf", group=current_group())
    root.component += file
    root.component.deploy()
    assert not os.chown.called


@patch("os.chown")
@patch("os.stat")
def test_group_calls_chown(chown, stat, root):
    os.stat.return_value = Mock()
    os.stat.return_value.st_gid = 0
    os.stat.return_value.st_mode = 0
    file = File("asdf", group=current_group(), content="")
    root.component += file
    root.component.deploy()
    assert os.chown.called


def test_group_is_configurable_when_group_doesnt_exist_yet(root):
    file = File("asdf", group="foobar", content="")
    root.component += file


def test_purge_globs_and_deletes_tree(root):
    os.mkdir("work/mycomponent/source")
    open("work/mycomponent/source/one", "w").close()
    open("work/mycomponent/source/two", "w").close()
    root.component += Purge("sourc*")
    root.component.deploy()
    assert sorted(os.listdir("work/mycomponent")) == []


def test_purge_globs_and_deletes_files(root):
    os.mkdir("work/mycomponent/source-one")
    open("work/mycomponent/source-two", "w").close()
    root.component += Purge("sourc*")
    root.component.deploy()
    assert sorted(os.listdir("work/mycomponent")) == []


def test_syncdirectory_needs_update_on_nonexisting_target(root):
    os.mkdir("work/mycomponent/existing_dir")
    open("work/mycomponent/existing_dir/test_file", "w").close()
    with pytest.raises(batou.UpdateNeeded):
        sd = SyncDirectory("non_existing_dir", source="existing_dir")
        sd.verify()


def test_directory_passes_args_to_syncdirectory(root):
    d = Directory(
        "target", source="source", verify_opts="-abc", sync_opts="-xyz"
    )
    d.prepare(root.component)
    sd = d._
    assert isinstance(sd, SyncDirectory)
    assert sd.verify_opts == "-abc"
    assert sd.sync_opts == "-xyz"



# This code must not cause non-stdlib imports to support self-bootstrapping.
import os
import os.path
import socket
import traceback
from typing import List, Optional

import jinja2

from ._output import output

with open(os.path.dirname(__file__) + "/version.txt") as f:
    __version__ = f.read().strip()

# Configure `remote-pdb` to be used with `breakpoint()` in Python 3.7+:
os.environ["PYTHONBREAKPOINT"] = "remote_pdb.set_trace"
if not os.environ.get("REMOTE_PDB_HOST", None):
    os.environ["REMOTE_PDB_HOST"] = "127.0.0.1"
if not os.environ.get("REMOTE_PDB_PORT", None):
    os.environ["REMOTE_PDB_PORT"] = "4444"


def prepare_error(error):
    return f"{error.__class__.__name__}: {error}"


def prepare_traceback(tb):
    from batou import component, environment

    stack = traceback.extract_tb(tb)
    while stack:
        # Delete remoting-internal stack frames.'
        line = stack.pop(0)
        if line[0] in [
            "<string>",
            "<remote exec>",
            environment.__file__.rstrip("c"),
            component.__file__.rstrip("c"),
        ]:
            continue
        stack.insert(0, line)
        break
    if not stack:
        return "<no-non-remote-internal-traceback-lines-found>"
    return "".join(traceback.format_list(stack))


class ReportingException(Exception):
    """Exceptions that support user-readable reporting."""

    affected_hostname: Optional[str]

    def __str__(self):
        raise NotImplementedError()

    def report(self):
        raise NotImplementedError()

    def should_merge(self, other):
        """
        checks, wether two exceptions have the same type as well as data
        and as such, should be merged into one exception.
        """
        # generic check:
        # the other class should have: the same class as self, in both directions
        if type(other) != type(self):
            return False
        # now both should have the same attributes
        # compare by all attributes except: .affected_hostname
        dict_self = self.__dict__.copy()
        dict_self.pop("affected_hostname", None)
        dict_other = other.__dict__.copy()
        dict_other.pop("affected_hostname", None)
        return dict_self == dict_other

    @classmethod
    def merge(cls, selfs):
        """Merge multiple instances of this exception."""
        # remove and collect .affected_hostname from all exceptions
        if hasattr(selfs[0], "affected_hostname"):
            hostnames = set()
            for self in selfs:
                hostnames.add(self.affected_hostname)
        else:
            hostnames = None
        # create a new exception with the merged attributes
        first_self_dict = selfs[0].__dict__.copy()
        first_self_dict.pop("affected_hostname", None)
        new_exception = cls()
        new_exception.__dict__.update(first_self_dict)
        new_exception.affected_hostname = "<HOSTNAME>"
        return (new_exception, hostnames)


class FileLockedError(ReportingException):
    """A file is already locked and we do not want to block."""

    filename: str

    @classmethod
    def from_context(cls, filename):
        self = cls()
        self.filename = filename
        return self

    def __str__(self):
        return "File already locked: {}".format(self.filename)

    def report(self):
        output.error(str(self))


class GPGCallError(ReportingException):
    """There was an error calling GPG on encrypted file."""

    command: str
    exitcode: str
    output: str

    @classmethod
    def from_context(cls, command, exitcode, output):
        self = cls()
        self.command = " ".join(command)
        self.exitcode = str(exitcode)
        self.output = output.decode("ascii", errors="replace")
        return self

    def __str__(self):
        return (
            f"Exitcode {self.exitcode} while calling: "
            f"{self.command}\n{self.output}"
        )

    def report(self):
        output.error("Error while calling GPG")
        output.tabular("command", self.command, red=True)
        output.tabular("exit code", self.exitcode)
        output.tabular("message", self.output, separator=":\n")


class AgeCallError(ReportingException):
    """There was an error calling age on encrypted file."""

    command: str
    exitcode: str
    output: str

    @classmethod
    def from_context(cls, command, exitcode, output):
        self = cls()
        self.command = " ".join(command)
        self.exitcode = str(exitcode)
        self.output = output.decode("ascii", errors="replace")
        return self

    def __str__(self):
        return (
            f"Exitcode {self.exitcode} while calling: "
            f"{self.command}\n{self.output}"
        )

    def report(self):
        output.error("Error while calling age")
        output.tabular("command", self.command, red=True)
        output.tabular("exit code", self.exitcode)
        output.tabular("message", self.output, separator=":\n")


class GetAddressInfoError(ReportingException, socket.gaierror):
    """There was an error calling getaddrinfo."""

    hostname: str
    error: str

    @classmethod
    def from_context(cls, hostname, error):
        self = cls()
        self.hostname = hostname
        self.error = error
        return self

    def __str__(self):
        return f"Error while resolving {self.hostname}: {self.error}"

    def report(self):
        output.error("Error while resolving hostname")
        output.tabular("hostname", self.hostname, red=True)
        output.tabular("message", self.error, separator=":\n")


class UpdateNeeded(AssertionError):
    """A component requires an update."""


class ConfigurationError(ReportingException):
    """Indicates that an environment could not be configured successfully."""

    message: str
    has_component: bool
    component_root_name: Optional[str]

    @property
    def sort_key(self):
        return (0, self.message)

    @classmethod
    def from_context(cls, message, component=None):
        self = cls()
        self.message = message
        self.has_component = component is not None
        self.component_root_name = component.root.name if component else None
        self.affected_hostname = component.root.host.name if component else None
        return self

    def __str__(self):
        return str(self.message)

    def report(self):
        message = self.message
        if self.has_component:
            message = "{}: {}".format(
                self.component_root_name,
                message,
            )
        output.error(message)


class AttributeExpansionError(ConfigurationError):
    """An override attribute could not be expanded properly."""

    component_breadcrumbs: str
    value_repr: str
    error_str: str
    key: str

    @property
    def sort_key(self):
        return (
            1,
            self.affected_hostname,
            self.component_breadcrumbs,
            self.key,
        )

    @classmethod
    def from_context(cls, component, key, value, error):
        self = cls()
        self.affected_hostname = component.root.host.name
        self.component_breadcrumbs = component._breadcrumbs
        self.value_repr = repr(value)
        self.error = prepare_error(error)
        self.key = key
        return self

    def __str__(self):
        return "Error while expanding attribute: " + self.error

    def report(self):
        output.error("Error while expanding attribute:")
        output.tabular(
            "Message",
            self.error,
            red=True,
        )
        output.tabular(
            "Attribute",
            "{}.{}".format(self.component_breadcrumbs, self.key),
            red=True,
        )


class ConversionError(ConfigurationError):
    """An override attribute could not be converted properly."""

    component_breadcrumbs: str
    conversion_name: str
    value_repr: str
    error_str: str
    key: str

    @property
    def sort_key(self):
        return (
            1,
            self.affected_hostname,
            self.component_breadcrumbs,
            self.key,
        )

    @classmethod
    def from_context(cls, component, key, value, conversion, error):
        self = cls()
        self.affected_hostname = component.root.host.name
        self.component_breadcrumbs = component._breadcrumbs
        self.conversion_name = conversion.__name__
        self.value_repr = repr(value)
        self.error_str = str(error)
        self.key = key
        return self

    def __str__(self):
        return self.error_str

    def report(self):
        output.error(self.error_str)
        output.tabular(
            "Attribute",
            "{}.{}".format(self.component_breadcrumbs, self.key),
            red=True,
        )
        output.tabular(
            "Conversion",
            "{}({})".format(self.conversion_name, self.value_repr),
            red=True,
        )
        # TODO provide traceback in debug output
        # provide file, line number, excerpt of attribute definition
        # see: https://github.com/flyingcircusio/batou/issues/316


class SilentConfigurationError(Exception):
    """These are exceptions that will be reported by other exceptions.

    They basically only influence control flow during configuration and
    are manually placed to avoid double reporting.

    """


class MissingOverrideAttributes(ConfigurationError):
    component_breadcrumbs: str
    attributes: List[str]

    @property
    def sort_key(self):
        return (3, self.affected_hostname, self.component_breadcrumbs)

    @classmethod
    def from_context(cls, component, attributes):
        self = cls()
        self.affected_hostname = component.root.host.name
        self.component_breadcrumbs = component._breadcrumbs
        self.attributes = attributes
        return self

    def __str__(self):
        return "Overrides for undefined attributes " + ",".join(self.attributes)

    def report(self):
        output.error("Overrides for undefined attributes")
        output.tabular("Component", self.component_breadcrumbs, red=True)
        output.tabular("Attributes", ", ".join(self.attributes), red=True)

        # TODO point to specific line in secrets or environments
        # cfg file and show context


class DuplicateComponent(ConfigurationError):
    a_name: str
    a_filename: str
    b_filename: str

    @property
    def sort_key(self):
        return (2, self.a_name)

    @classmethod
    def from_context(cls, a, b):
        self = cls()
        self.a_name = a.name
        self.a_filename = a.filename
        self.b_filename = b.filename
        return self

    def __str__(self):
        return "Duplicate component: " + self.a_name

    def report(self):
        output.error('Duplicate component "{}"'.format(self.a_name))
        output.tabular("Occurrence", self.a_filename)
        output.tabular("Occurrence", self.b_filename)


class DuplicateHostMapping(ConfigurationError):
    affected_hostname: str
    a: str
    b: str

    @property
    def sort_key(self):
        return (3, self.affected_hostname, self.a, self.b)

    @classmethod
    def from_context(cls, hostname, a, b):
        self = cls()
        self.affected_hostname = hostname
        self.a = a
        self.b = b
        return self

    def __str__(self):
        return "Duplicate host mapping: " + self.affected_hostname

    def report(self):
        output.error(
            'Duplicate host mapping "{}"'.format(self.affected_hostname)
        )
        output.tabular("Mapping 1: ", self.a)
        output.tabular("Mapping 2: ", self.b)


class UnknownComponentConfigurationError(ConfigurationError):
    """An unknown error occured while configuring a component."""

    @property
    def sort_key(self):
        return (4, self.root_host_name, 2)

    @classmethod
    def from_context(cls, root, exception, tb):
        self = cls()
        self.root_name = root.name
        self.root_host_name = root.host.name
        self.exception_repr = prepare_error(exception)
        stack = traceback.extract_tb(tb)
        from batou import component, environment

        self.traceback = prepare_traceback(tb)
        return self

    def __str__(self):
        return self.exception_repr

    def report(self):
        output.error(self.exception_repr)
        output.annotate(
            "This /might/ be a batou bug. Please consider reporting it.\n",
            red=True,
        )
        output.tabular("Host", self.root_host_name, red=True)
        output.tabular("Component", self.root_name + "\n", red=True)
        output.annotate(
            "Traceback (simplified, most recent call last):", red=True
        )
        output.annotate(self.traceback, red=True)


class UnusedResources(ConfigurationError):
    """Some provided resources were never used."""

    sort_key = (5, "unused")

    @classmethod
    def from_context(cls, resources):
        self = cls()
        self.unused_resources = []
        for key in sorted(resources.keys()):
            for component, value in resources[key].items():
                self.unused_resources.append((key, component.name, str(value)))
        return self

    def __str__(self):
        return "Unused provided resources"

    def report(self):
        output.error("Unused provided resources")
        for key, component, value in self.unused_resources:
            output.line(
                '    Resource "{}" provided by {} with value {}'.format(
                    key, component, value
                ),
                red=True,
            )


class UnusedComponentsInitialized(ConfigurationError):
    """Some components were initialized but never used."""

    sort_key = (5, "unused")

    @classmethod
    def from_context(cls, components, root):
        self = cls()
        self.unused_components = []
        self.breadcrumbs = []
        self.init_file_paths = []
        self.init_line_numbers = []
        for component in components:
            self.unused_components.append(repr(component.__class__.__name__))
            self.breadcrumbs.append(component._init_breadcrumbs)
            self.init_file_paths.append(component._init_file_path)
            self.init_line_numbers.append(component._init_line_number)
        self.root_name = root.name
        return self

    def __str__(self):
        out_str = "Some components were initialized but never added to the environment:"
        for i, component in enumerate(self.unused_components):
            out_str += f"\n    {component}: {' -> '.join(self.breadcrumbs[i])}"
            out_str += f"\n        initialized in {self.init_file_paths[i]}:{self.init_line_numbers[i]}"
        out_str += f"\nRoot: {self.root_name}"
        out_str += f"\nAdd the components to the environment using `self += component`."
        return out_str

    def report(self):
        output.error(
            f"Some components were initialized but never added to the environment:"
        )
        for i, component in enumerate(self.unused_components):
            output.line(
                f"    {component}: {' -> '.join(self.breadcrumbs[i])}", red=True
            )
            output.line(
                f"        initialized in {self.init_file_paths[i]}:{self.init_line_numbers[i]}",
                red=True,
            )
        output.line(
            f"Add the components to the environment using `self += component`.",
            red=True,
        )
        output.tabular("Root", self.root_name, red=True)


class ComponentWithUpdateWithoutVerify(ConfigurationError):
    """Some components have an update method but no verify method."""

    sort_key = (5, "without_verify")

    @classmethod
    def from_context(cls, components, roots):
        self = cls()
        self.components = []
        for component in components:
            self.components.append(repr(component.__class__.__name__))
        self.roots = []
        for root in roots:
            self.roots.append(root.name)
        return self

    def __str__(self):
        out_str = "Some components have an update method but no verify method:"
        for idx, component in enumerate(self.components):
            out_str += f"\n    {component}"
            out_str += f"\nRoot: {self.roots[idx]}"
        out_str += f"\nThe update() method may not be called by batou if the verify() method is missing."
        return out_str

    def report(self):
        output.error(
            f"Some components have an update method but no verify method:"
        )
        for component in self.components:
            output.line(f"    {component}", red=True)
        output.tabular("Root", self.root_name, red=True)
        output.line(
            f"The update() method may not be called by batou if the verify() method is missing.",
            red=True,
        )


class UnsatisfiedResources(ConfigurationError):
    """Some required resources were never provided."""

    sort_key = (6, "unsatisfied")

    @classmethod
    def from_context(cls, resources):
        self = cls()
        self.unsatisfied_resources = []
        for (key, host), res_for_key in sorted(resources.items()):
            self.unsatisfied_resources.append(
                (key, host, [r.name for r in res_for_key])
            )
        return self

    def __str__(self):
        return "Unsatisfied resource requirements"

    def report(self):
        output.error("Unsatisfied resource requirements")
        for key, host, resources in self.unsatisfied_resources:
            if host is None:
                msg = (
                    f'    Resource "{key}" required by '
                    f'{",".join(resources)}'
                )

            else:
                msg = (
                    f'    Resource "{key}" on "{host}" required by '
                    f'{",".join(resources)}'
                )

            output.line(msg, red=True)


class MissingEnvironment(ConfigurationError):
    """The specified environment does not exist."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, environment):
        self = cls()
        self.environment_name = environment.name
        return self

    def __str__(self):
        return "Missing environment `{}`".format(self.environment_name)

    def report(self):
        output.error("Missing environment")
        output.tabular("Environment", self.environment_name, red=True)


class ComponentLoadingError(ConfigurationError):
    """The specified component file failed to load."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, filename, exception, tb):
        self = cls()
        self.filename = filename
        self.exception_str = str(exception)

        self.traceback = prepare_traceback(tb)
        return self

    def __str__(self):
        return "Failed loading component file " + self.filename

    def report(self):
        output.error("Failed loading component file")
        output.tabular("File", self.filename, red=True)
        output.tabular("Exception", self.exception_str, red=True)
        output.annotate(
            "Traceback (simplified, most recent call last):", red=True
        )
        output.annotate(self.traceback, red=True)

        # TODO provide traceback in debug output
        # see: https://github.com/flyingcircusio/batou/issues/316


class MissingComponent(ConfigurationError):
    """The specified environment does not exist."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, component_name, hostname):
        self = cls()
        self.component_name = component_name
        self.affected_hostname = hostname
        return self

    def __str__(self):
        return "Missing component: " + self.component_name

    def report(self):
        output.error("Missing component")
        output.tabular("Component", self.component_name, red=True)


class SuperfluousSection(ConfigurationError):
    """A superfluous section was found in the environment
    configuration file."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, section):
        self = cls()
        self.section = section
        return self

    def __str__(self):
        return "Superfluous section in environment: " + self.section

    def report(self):
        output.error("Superfluous section in environment configuration")
        output.tabular("Section", self.section, red=True)
        # TODO provide location and context in debug output
        # see: https://github.com/flyingcircusio/batou/issues/316


class SuperfluousComponentSection(ConfigurationError):
    """A component section was found in the environment
    but no associated component is known."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, component_name):
        self = cls()
        self.component_name = component_name
        return self

    def __str__(self):
        return (
            "Override section for unknown component found: "
            + self.component_name
        )

    def report(self):
        output.error("Override section for unknown component found")
        output.tabular("Component", self.component_name, red=True)
        # TODO provide traceback in debug output
        # see: https://github.com/flyingcircusio/batou/issues/316


class SuperfluousSecretsSection(ConfigurationError):
    """A component section was found in the secrets
    but no associated component is known."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, component_name):
        self = cls()
        self.component_name = component_name
        return self

    def __str__(self):
        return (
            "Secrets section for unknown component found: "
            + self.component_name
        )

    def report(self):
        output.error("Secrets section for unknown component found")
        output.tabular("Component", self.component_name, red=True)
        # TODO provide traceback in debug output
        # see: https://github.com/flyingcircusio/batou/issues/316


class UnknownHostSecretsSection(ConfigurationError):
    """A host section was found in the secrets
    but no associated host is known."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, hostname):
        self = cls()
        self.hostname = hostname
        return self

    def __str__(self):
        return "Secrets section for unknown host found: " + self.hostname

    def report(self):
        output.error("Secrets section for unknown host found")
        output.tabular("Host", self.hostname, red=True)


class DuplicateOverride(ConfigurationError):
    """An override for a component attribute was found both in the secrets and
    in the environment configuration."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, component_name, attribute):
        self = cls()
        self.component_name = component_name
        self.attribute = attribute
        return self

    def __str__(self):
        return (
            f"A value {self.component_name}.{self.attribute} is defined both in"
            " environment and secrets."
        )

    def report(self):
        output.error("Attribute override found both in environment and secrets")
        output.tabular("Component", self.component_name, red=True)
        output.tabular("Attribute", self.attribute, red=True)
        # TODO provide traceback in debug output
        # see: https://github.com/flyingcircusio/batou/issues/316


class DuplicateSecretsComponentAttribute(ConfigurationError):
    """A component attribute was specified multiple times in the secrets."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, component_name, attribute):
        self = cls()
        self.component_name = component_name
        self.attribute = attribute
        return self

    def __str__(self):
        return (
            f"A value {self.component_name}.{self.attribute} is defined multiple"
            " times in secrets."
        )

    def report(self):
        output.error("Attribute defined multiple times in secrets")
        output.tabular("Component", self.component_name, red=True)
        output.tabular("Attribute", self.attribute, red=True)
        # TODO provide traceback in debug output
        # see: https://github.com/flyingcircusio/batou/issues/316


class CycleErrorDetected(ConfigurationError):
    """We think we found a cycle in the component dependencies."""

    sort_key = (99,)

    @classmethod
    def from_context(cls, error):
        self = cls()
        self.error_str = str(error)
        return self

    def report(self):
        output.error("Found dependency cycle")
        output.annotate(self.error_str, red=True)
        # TODO provide traceback in debug output
        # see: https://github.com/flyingcircusio/batou/issues/316

    def __str__(self):
        return "Found dependency cycle: " + self.error_str


class NonConvergingWorkingSet(ConfigurationError):
    """A working set did not converge."""

    sort_key = (100,)

    @classmethod
    def from_context(cls, roots):
        self = cls()
        self.roots_len = len(roots)
        self.root_names = ", ".join(sorted([r.name for r in roots]))
        return self

    def __str__(self):
        return "There are unconfigured components remaining."

    def report(self):
        # TODO show this last or first, but not in the middle
        # of everything
        output.error(
            "{} remaining unconfigured component(s): {}".format(
                self.roots_len, self.root_names
            )
        )
        # TODO show all incl. their host name in -vv or so


class DeploymentError(ReportingException):
    """Indicates that a deployment failed.."""

    sort_key = (100,)

    def __str__(self):
        return "The deployment encountered an error."

    def report(self):
        pass


class RepositoryDifferentError(DeploymentError):
    """The repository on the remote side is different."""

    sort_key = (150,)

    @classmethod
    def from_context(cls, local, remote):
        self = cls()
        self.local = local
        self.remote = remote
        return self

    def __str__(self):
        return "Remote repository has diverged. Wrong branch?"

    def report(self):
        output.error(
            "The remote working copy is based on a different revision. "
            "Maybe you tried to deploy from the wrong branch."
        )
        output.tabular("Local", self.local, red=True)
        output.tabular("Remote", self.remote, red=True)


class DuplicateHostError(ConfigurationError):
    @property
    def sort_key(self):
        return (0,)

    @classmethod
    def from_context(cls, hostname):
        self = cls()
        self.affected_hostname = hostname
        return self

    def __str__(self):
        return "Duplicate host: " + self.affected_hostname

    def report(self):
        output.error(
            "Duplicate definition of host: {}".format(self.affected_hostname)
        )


class InvalidIPAddressError(ConfigurationError):
    sort_key = (0,)

    @classmethod
    def from_context(cls, address):
        self = cls()
        self.address = address
        return self

    def __str__(self):
        return "Not a valid IP address: " + self.address

    def report(self):
        output.error("Not a valid IP address: {}".format(self.address))


class IPAddressConfigurationError(ConfigurationError):
    """An IP address family was accessed but not configured."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, address, kind):
        self = cls()
        self.address = address
        self.kind = kind
        return self

    def __str__(self):
        return (
            f"Trying to access address family IPv{self.kind} which is not"
            f" configured for {self.address}."
        )

    def report(self):
        output.error(str(self))

        output.tabular(
            "Hint",
            f"Use `require_v{self.kind}=True` when instantiating the Address"
            " object.",
            red=True,
        )

        # TODO provide traceback/line numbers/excerpt
        # see: https://github.com/flyingcircusio/batou/issues/316


class TemplatingError(ReportingException):
    """An error occured while rendering a template."""

    sort_key = (0,)

    @classmethod
    def from_context(cls, exception, template_identifier):
        self = cls()
        self.exception_str = prepare_error(exception)
        self.template_identifier = template_identifier
        # if exception is instance of jinja2.TemplateSyntaxError
        # there is some magic in jinja2 that makes the __str__ method
        # less capable, if exception.translated is set to True
        if isinstance(exception, jinja2.TemplateSyntaxError):
            exception.translated = False
            self.exception_str = str(exception)
            exception.translated = True
        return self

    def __str__(self):
        return f"An error occured while rendering a template ({self.template_identifier}): {self.exception_str}"

    def report(self):
        output.error(str(self))



import configparser
import os
import os.path
import pathlib
import shutil
import tempfile

import mock
import pytest

import batou
from batou.secrets import GPGSecretProvider
from batou.secrets.encryption import (
    AGEEncryptedFile,
    DiffableAGEEncryptedFile,
    EncryptedFile,
    GPGEncryptedFile,
)

FIXTURE = pathlib.Path(__file__).parent / "fixture"
cleartext_file = FIXTURE / "cleartext.cfg"
FIXTURE_ENCRYPTED_CONFIG = FIXTURE / "encrypted.cfg.gpg"


@pytest.fixture(scope="function")
def encrypted_file(tmpdir):
    """Provide a temporary copy of the encrypted confi."""
    return pathlib.Path(shutil.copy(FIXTURE_ENCRYPTED_CONFIG, tmpdir))


@pytest.fixture(scope="session", autouse=True)
def cleanup_gpg_sockets():
    yield
    for path in [
        "S.dirmngr",
        "S.gpg-agent",
        "S.gpg-agent.browser",
        "S.gpg-agent.extra",
        "S.gpg-agent.ssh",
    ]:
        try:
            (FIXTURE / "gnupg" / path).unlink()
        except OSError:
            pass


def test_error_message_no_gpg_found(encrypted_file):
    c = GPGEncryptedFile(encrypted_file)
    OLD_GPG_BINARY_CANDIDATES = GPGEncryptedFile.GPG_BINARY_CANDIDATES
    GPGEncryptedFile.GPG_BINARY_CANDIDATES = ["foobarasdf-54875982"]
    GPGEncryptedFile._gpg = None
    with pytest.raises(RuntimeError) as e:
        c.gpg()
    assert e.value.args[0] == (
        "Could not find gpg binary. Is GPG installed? I tried looking for: "
        "`foobarasdf-54875982`"
    )
    GPGEncryptedFile.GPG_BINARY_CANDIDATES = OLD_GPG_BINARY_CANDIDATES


def test_error_message_no_age_found(encrypted_file):
    c = AGEEncryptedFile(encrypted_file)
    OLD_AGE_BINARY_CANDIDATES = AGEEncryptedFile.AGE_BINARY_CANDIDATES
    AGEEncryptedFile.AGE_BINARY_CANDIDATES = ["foobarasdf-54875982"]
    AGEEncryptedFile._age = None
    with pytest.raises(RuntimeError) as e:
        c.age()
    assert e.value.args[0] == (
        "Could not find age binary. Is age installed? I tried looking for: "
        "`foobarasdf-54875982`"
    )
    AGEEncryptedFile.AGE_BINARY_CANDIDATES = OLD_AGE_BINARY_CANDIDATES


def test_decrypt(encrypted_file):
    with GPGEncryptedFile(encrypted_file) as secret:
        with open(cleartext_file) as cleartext:
            assert cleartext.read().strip() == secret.cleartext.strip()


def test_decrypt_missing_key(monkeypatch, encrypted_file):
    monkeypatch.setitem(os.environ, "GNUPGHOME", "/tmp")

    with pytest.raises(batou.GPGCallError):
        with GPGEncryptedFile(encrypted_file) as secret:
            secret.cleartext


def test_write_should_fail_unless_write_locked(encrypted_file):
    with GPGEncryptedFile(encrypted_file) as secret:
        secret.cleartext
        with pytest.raises(RuntimeError):
            secret.write(b"", [])


def test_open_nonexistent_file_for_read_should_fail():
    with pytest.raises(IOError):
        with GPGEncryptedFile(pathlib.Path("/no/such/file")) as secret:
            secret.cleartext


def test_open_nonexistent_file_for_write_should_create_empty_lock_file():
    tf = tempfile.NamedTemporaryFile(prefix="new_encrypted.")
    tf.close()  # deletes file
    encrypted = GPGEncryptedFile(pathlib.Path(tf.name), writeable=True)
    with encrypted as secrets:
        assert secrets.cleartext == ""
        # The file exists, because we set the write lock
        assert os.path.exists(tf.name)
    # When exiting a file without writing, the lock file is removed
    assert not os.path.exists(tf.name)


def test_write(encrypted_file):
    # encrypted = EncryptedConfigFile(encrypted_file, write_lock=True)
    encrypted = GPGEncryptedFile(pathlib.Path(encrypted_file), writeable=True)
    with encrypted as secrets:
        secrets.write(
            b"""\
[batou]
members = batou
[asdf]
x = 1
""",
            ["batou"],
        )

    assert FIXTURE_ENCRYPTED_CONFIG.read_bytes() != encrypted_file.read_bytes()
    assert 0 != encrypted_file.stat().st_size


def test_write_fails_if_recipient_key_is_missing_keeps_old_file(encrypted_file):
    encrypted = GPGEncryptedFile(pathlib.Path(encrypted_file), writeable=True)
    with encrypted as secrets:
        with pytest.raises(batou.GPGCallError):
            secrets.write(
                b"""\
[batou]
members = foobar@example.com
[asdf]
x = 1
""",
                ["foobar@example.com"],
            )

    assert encrypted_file.read_bytes() == FIXTURE_ENCRYPTED_CONFIG.read_bytes()


def test_write_and_read_age_diffable(encrypted_file):
    encrypted = DiffableAGEEncryptedFile(
        pathlib.Path(encrypted_file), writeable=True
    )
    content = b"""\
[batou]
members = ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIACZ8++sQADp8fztgumfw2i+WSgzMHB7MgSpkM2y5pHi batou-ci-test-key

[asdf]
value = This is a very long
    multiline string
    that should be encrypted
    and decrypted correctly
"""
    # parse content as config to ensure that the content is valid
    config = configparser.ConfigParser()
    config.read_string(content.decode("utf-8"))

    with encrypted as secrets:
        secrets.write(
            content,
            [
                "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIACZ8++sQADp8fztgumfw2i+WSgzMHB7MgSpkM2y5pHi batou-ci-test-key"
            ],
        )

    assert content != encrypted_file.read_bytes()
    assert 0 != encrypted_file.stat().st_size

    with DiffableAGEEncryptedFile(encrypted_file) as secrets:
        # we want the multiline string to be read correctly
        secret_config = configparser.ConfigParser()
        secret_config.read_string(secrets.cleartext)
        assert config["asdf"]["value"] == secret_config["asdf"]["value"]



from batou.component import Component
from batou.lib.file import File


class ZEO(Component):
    def configure(self):
        self.provide("zeo", "127.0.0.1:9000")
        # As reported in #11183 we had a regression that caused the
        # sub-component to remove the provided values accidentally.
        self += File("asdf", content="asdf")



from batou.utils import get_output


def migrate(output):
    candidates = get_output(
        'grep "Adddress" * -nIr | egrep -v "import"', "<no candidates found>"
    )
    output(
        "Address objects now only resolve IPv4 addresses by default",
        f"""

        If you use IPv6 then you have to enable it explicitly by setting
        `require_v6=True`. If you want the old behaviour you can use
        `require_v6='optional'`.

        You can also set new defaults per environment settings which allows
        you to return to the previous behaviour of IPv6 being optional without
        touching component code:

        [environment]
        require_v4 = True
        require_v6 = optional

        Candidates:

        {candidates}
        """,
        "manual",
    )



import os.path

from batou.lib.appenv import AppEnv


def test_simple_appenv(root):
    with open("requirements.lock", "w") as f:
        # I hate using a real package and a real index here ...
        f.write("six==1.14.0\n")

    appenv = AppEnv("3")
    root.component += appenv
    root.component.deploy()

    assert os.path.exists(os.path.join(root.component.workdir, "bin"))
    assert os.path.exists(
        os.path.join(root.component.workdir, "bin", "python3")
    )
    assert os.path.exists(
        os.path.join(root.component.workdir, ".appenv", "current")
    )

    hashes = os.listdir(os.path.join(root.component.workdir, ".appenv"))
    assert len(hashes) == 2
    first_hash = list(set(hashes) - set(["current"]))[0]

    # Running it twice doesn't change anything
    appenv.prepare(root.component)
    root.component.deploy()

    assert os.path.exists(os.path.join(root.component.workdir, "bin"))
    assert os.path.exists(
        os.path.join(root.component.workdir, "bin", "python3")
    )
    assert os.path.exists(
        os.path.join(root.component.workdir, ".appenv", "current")
    )

    hashes2 = os.listdir(os.path.join(root.component.workdir, ".appenv"))
    assert hashes2 == hashes
    assert len(hashes2) == 2
    # Changing the requirements does change something:
    with open("requirements.lock", "w") as f:
        # I hate using a real package and a real index here ...
        f.write("# comment\nsix==1.14.0\n")

    appenv.prepare(root.component)
    root.component.deploy()

    assert os.path.exists(os.path.join(root.component.workdir, "bin"))
    assert os.path.exists(
        os.path.join(root.component.workdir, "bin", "python3")
    )
    assert os.path.exists(
        os.path.join(root.component.workdir, ".appenv", "current")
    )

    hashes3 = os.listdir(os.path.join(root.component.workdir, ".appenv"))
    assert hashes3 != hashes

    assert len(hashes3) == 3
    second_hash = list(set(hashes3) - set(["current", first_hash]))[0]
    assert first_hash == appenv.last_env_hash
    assert second_hash == appenv.env_hash

    # Changing the requirements again will remove `first_hash`
    with open("requirements.lock", "w") as f:
        # I hate using a real package and a real index here ...
        f.write("# another comment\nsix==1.14.0\n")

    appenv.prepare(root.component)
    root.component.deploy()

    hashes4 = os.listdir(os.path.join(root.component.workdir, ".appenv"))
    assert len(hashes4) == 3

    assert "current" in hashes4
    assert second_hash in hashes4
    assert first_hash not in hashes4


def test_appenv_custom_pip_version(root):
    pip_version = "24.0"
    with open("requirements.lock", "w") as f:
        # I hate using a real package and a real index here ...
        f.write("six==1.14.0\n")

    appenv = AppEnv("3", pip_version=pip_version)
    root.component += appenv
    root.component.deploy()
    pip = appenv.cmd(
        os.path.join(
            root.component.workdir, "{{component.env_dir}}/bin/pip --version"
        )
    )
    assert f"pip {pip_version}" in pip[0]



import os

from batou import ConfigurationError, UpdateNeeded
from batou.component import (
    Attribute,
    Component,
    ConfigString,
    HookComponent,
    platform,
)
from batou.lib.file import File


class CronJob(HookComponent):

    namevar = "command"
    key = "batou.lib.cron:CronJob"

    args = ""
    timing = None
    logger = None

    def format(self):
        if self.timing is None:
            raise ValueError(
                "Required timing value missing from cron job %r." % self.command
            )
        line = self.expand(
            "{{component.timing}} {{component.command}} {{component.args}}"
        )
        if self.logger:
            line += self.expand(" 2>&1 | logger -t {{component.logger}}")
        return line


def ignore_comments(data):
    lines = data.splitlines()
    lines = [x for x in lines if not x.startswith(b"#")]
    return b"\n".join(lines)


class CronTab(Component):

    crontab_template = os.path.join(
        os.path.dirname(__file__), "resources", "crontab"
    )
    mailto = Attribute(default=None)
    purge = False

    # Dict of additional environment variables
    env = Attribute("literal", default=ConfigString("{}"))

    def configure(self):
        self.jobs = self.require(CronJob.key, host=self.host, strict=False)
        if self.purge and self.jobs:
            raise ConfigurationError.from_context(
                "Found cron jobs, but expecting an empty crontab."
            )
        elif not self.purge and not self.jobs:
            raise ConfigurationError.from_context("No cron jobs found.", self)
        self.jobs.sort(key=lambda job: job.command + " " + job.args)
        self.crontab = File("crontab", source=self.crontab_template)
        self += self.crontab


class PurgeCronTab(Component):
    def configure(self):
        self += CronTab(purge=True)


class InstallCrontab(Component):
    def configure(self):
        self.crontab = self.parent.crontab

    def verify(self):
        try:
            current, _ = self.cmd("crontab -l", encoding=None)
        except Exception:
            current = ""
        current = ignore_comments(current)
        new = ignore_comments(self.crontab.content)
        if new != current:
            raise UpdateNeeded()

    def update(self):
        self.cmd(self.expand("crontab {{component.crontab.path}}"))


@platform("gocept.net", CronTab)
class FCInstallCrontab(InstallCrontab):
    pass


@platform("debian", CronTab)
class DebianInstallCrontab(InstallCrontab):
    pass



import os
import pathlib
import re
import urllib.request
from typing import TYPE_CHECKING, Dict, List, Optional, Set

from configupdater import ConfigUpdater

from batou import (
    DuplicateOverride,
    DuplicateSecretsComponentAttribute,
    SuperfluousSecretsSection,
    UnknownHostSecretsSection,
)
from batou._output import output

from .encryption import (
    AGEEncryptedFile,
    DiffableAGEEncryptedFile,
    EncryptedFile,
    GPGEncryptedFile,
    NoBackingEncryptedFile,
    debug,
)

if TYPE_CHECKING:
    from batou.environment import Environment


class SecretProvider:
    @classmethod
    def from_environment(cls, environment: "Environment") -> "SecretProvider":
        """
        Inspects the environment to determine which secret provider to use.
        Returns an instance of the secret provider.
        """
        environment_path = (
            pathlib.Path(environment.base_dir)
            / "environments"
            / str(environment.name)
        )
        secret_provider_candidates: List[SecretProvider] = []

        # check if environment has a environment.cfg file. If not, it does not
        # exist.

        if not (environment_path / "environment.cfg").exists():
            raise ValueError(f"Environment {environment.name} does not exist.")

        output.annotate(
            f"Looking for secrets for environment {environment.name}.",
            debug=True,
        )

        # There should be one file called "secrets.cfg.<provider>".
        # If there is more than one, we raise an error.

        config_files = list(environment_path.glob("secrets.cfg.*"))
        if len(config_files) == 0:
            output.annotate(
                f"No secrets found for environment {environment.name}.",
                debug=True,
            )
            return NoSecretProvider(environment)
        if len(config_files) > 1:
            raise ValueError(
                f"Multiple secret providers found for environment {environment.name}.",
                f"Candidates: {config_files}. Cannot continue.",
            )

        extension = config_files[0].suffix

        if extension == ".age":
            output.annotate(
                f"Found age secrets for environment {environment.name}.",
                debug=True,
            )
            secret_provider_candidates.append(AGESecretProvider(environment))
        elif extension == ".gpg":
            output.annotate(
                f"Found gpg secrets for environment {environment.name}.",
                debug=True,
            )
            secret_provider_candidates.append(GPGSecretProvider(environment))
        elif extension == ".age-diffable":
            output.annotate(
                f"Found age-diffable secrets for environment {environment.name}.",
                debug=True,
            )
            secret_provider_candidates.append(
                DiffableAGESecretProvider(environment)
            )
        else:
            raise ValueError(
                f"Invalid secret provider {extension}.",
            )

        return secret_provider_candidates[0]

    def __init__(self, environment: "Environment"):
        self.environment = environment

    @property
    def config(self) -> ConfigUpdater:
        raise NotImplementedError(
            "Cannot read encrypted config where no secret provider is configured."
        )

    def read(self) -> "SecretBlob":
        """
        Read the secrets for the environment and return a SecretBlob.
        """
        raise NotImplementedError("read() not implemented.")

    def read_secret_files(self) -> Dict[str, bytes]:
        """
        Read the secret files for the environment and return a dict of
        filename -> content.
        """
        raise NotImplementedError("read_secret_files() not implemented.")

    def inject_secrets(self):
        """
        Inject secrets into the environment.
        """
        output.annotate(
            f"Injecting secrets for environment {self.environment.name}.",
            debug=True,
        )
        secret_blob = self.read()
        for hostname in secret_blob.host_data:
            if hostname not in self.environment.hosts:
                self.environment.exceptions.append(
                    UnknownHostSecretsSection.from_context(hostname)
                )
                continue
            for key, value in secret_blob.host_data[hostname].items():
                self.environment.hosts[hostname].data[key] = value

        for component_name in secret_blob.component_overrides:
            if component_name not in self.environment.components:
                self.environment.exceptions.append(
                    SuperfluousSecretsSection.from_context(component_name)
                )
            overrides = self.environment.overrides.setdefault(
                component_name, {}
            )
            for key, value in secret_blob.component_overrides[
                component_name
            ].items():
                keys_added = set()
                if key in keys_added:
                    self.environment.exceptions.append(
                        DuplicateSecretsComponentAttribute.from_context(
                            component_name, key
                        )
                    )
                elif key in overrides:
                    self.environment.exceptions.append(
                        DuplicateOverride.from_context(component_name, key)
                    )
                else:
                    overrides[key] = value
                    keys_added.add(key)

        self.environment.secret_data.update(secret_blob.secret_data)
        self.environment.secret_files.update(secret_blob.secret_files)

        output.annotate(
            f"Injected secrets for environment {self.environment.name}.",
            debug=True,
        )

    def summary(self):
        """
        Print a summary of the secrets.
        """
        raise NotImplementedError("summary() not implemented.")

    def edit(self, edit_file: Optional[str] = None) -> EncryptedFile:
        """
        Edit the secrets.
        """
        raise NotImplementedError("edit() not implemented.")

    def write_file(self, file: EncryptedFile, content: bytes):
        raise NotImplementedError("write_file() not implemented.")

    def write_config(self, content: bytes, force_reencrypt: bool = False):
        raise NotImplementedError("write_config() not implemented.")

    def write_config_new(self, content: bytes):
        raise NotImplementedError("write_config_new() not implemented.")

    def change_secret_provider(
        self,
        config: ConfigUpdater,
        old_secret_provider: "SecretProvider",
    ):
        new_secret_provider_str = config.get("batou", "secret_provider").value
        if new_secret_provider_str == "age":
            new_secret_provider = AGESecretProvider(self.environment)
        elif new_secret_provider_str == "gpg":
            new_secret_provider = GPGSecretProvider(self.environment)
        elif new_secret_provider_str == "age-diffable":
            new_secret_provider = DiffableAGESecretProvider(self.environment)
        else:
            raise ValueError(
                f"Invalid secret provider {new_secret_provider_str}.",
            )
        output.annotate(
            f"Changing secret provider from {old_secret_provider.secret_provider_str} to {new_secret_provider.secret_provider_str}."
        )
        new_secret_provider.write_config_new(str(config).encode("utf-8"))
        new_secret_provider.write_secret_files(
            old_secret_provider.read_secret_files()
        )
        self.environment.secret_provider = new_secret_provider
        old_secret_provider.purge(new_secret_provider.iter_secret_files())
        output.annotate(
            f"Secret provider changed from {old_secret_provider.secret_provider_str} to {new_secret_provider.secret_provider_str}."
        )

    def purge(self, except_files: Dict[str, EncryptedFile] = {}):
        raise NotImplementedError("purge() not implemented.")

    def _get_recipients(self) -> List[str]:
        raise NotImplementedError("_get_recipients() not implemented.")


class SecretBlob:
    def __init__(
        self,
        host_data: Dict[str, Dict[str, str]],
        component_overrides: Dict[str, Dict[str, str]],
        secret_data: Set[str],
        secret_files: Dict[str, str],
    ):
        """
        Holds the secrets for an environment.
        """
        self.host_data = host_data
        self.component_overrides = component_overrides
        self.secret_data = secret_data
        self.secret_files = secret_files


class NoSecretProvider(SecretProvider):
    secret_provider_str = "none"

    def __init__(self, environment: "Environment"):
        super().__init__(environment)

    def read(self):
        return SecretBlob({}, {}, set(), {})

    def summary(self):
        print("\tNo secrets found.")

    def edit(self, edit_file: Optional[str] = None):
        output.annotate(
            f"Editing secrets for environment {self.environment.name} without secret configuration.",
            debug=True,
        )
        if edit_file is not None:
            raise ValueError(
                "Cannot edit secret files for environment without secret configuration.",
            )
        return NoBackingEncryptedFile()

    def write_config(self, content: bytes, force_reencrypt: bool = False):
        output.annotate(
            f"Writing secrets configuration for environment {self.environment.name} without secret configuration.",
            debug=True,
        )
        self.change_secret_provider(
            ConfigUpdater().read_string(content.decode("utf-8")), self
        )

    def read_secret_files(self) -> Dict[str, bytes]:
        return {}

    def purge(self, except_files: Dict[str, EncryptedFile] = {}):
        pass


class ConfigFileSecretProvider(SecretProvider):
    config_file: EncryptedFile

    @property
    def config(self):
        return ConfigUpdater().read_string(self.config_file.cleartext)

    def read(self):
        host_data = {}  # associate hostnames with their data
        component_overrides = {}  # associate component names with their data
        secret_data = (
            set()
        )  # set of all secret data, used to check for secrets when outputting diffs
        with self.config_file:
            for section in self.config.sections():
                if section == "batou":
                    continue
                elif section.startswith("host:"):
                    hostname = section[len("host:") :]
                    host_data[hostname] = {}
                    for key, option in self.config[section].items():
                        key = key.replace("data-", "", 1)
                        host_data[hostname][key] = option.value
                else:
                    component_name = section.replace("component:", "")
                    component_overrides[component_name] = {}
                    for key, option in self.config[section].items():
                        component_overrides[component_name][key] = option.value
                        if option.value is not None:
                            secret_data.update(option.value.split())

        secret_files = self.read_secret_files()
        secret_files = {k: v.decode("utf-8") for k, v in secret_files.items()}

        for content in secret_files.values():
            secret_data.update(content.splitlines())

        return SecretBlob(
            host_data, component_overrides, secret_data, secret_files
        )

    def iter_secret_files(self, writeable=False) -> Dict[str, EncryptedFile]:
        raise NotImplementedError("iter_secret_files() not implemented.")

    def read_secret_files(self) -> Dict[str, bytes]:
        secret_files = {}
        for filename, file in self.iter_secret_files().items():
            with file:
                secret_files[filename] = file.decrypted
        return secret_files

    def write_secret_files(self, secret_files: Dict[str, bytes]):
        for name, content in secret_files.items():
            with self._get_file(name, writeable=True) as file:
                self.write_file(file, content)

    def purge(self, except_files: Dict[str, EncryptedFile] = {}):
        for name, file in self.iter_secret_files(writeable=True).items():
            if name not in except_files:
                file.delete()
        self.config_file.delete()

    def summary(self):
        print("\t members")
        with self.config_file:
            members = self.config.get("batou", "members")
            if members.value is not None:
                for member in re.split(r"(\n|,)+", members.value):
                    member = member.strip()
                    if member:
                        print(f"\t\t- {member}")
            else:
                print("\t\tUndefined behavior.")
            if not members:
                print("\t\t(none)")
            print("\t secret files")
            # Keys of self.f.files are strings, but self.path is pathlib.Path:
            files = self.read_secret_files().keys()
            files = sorted(files)
            for f in files:
                print("\t\t-", f)
            if not files:
                print("\t\t(none)")
            print()

    def edit(self, edit_file: Optional[str] = None):
        if edit_file is None:
            self.config_file.writeable = True
            return self.config_file
        else:
            return self._get_file(edit_file, writeable=True)

    def _get_file(self, name: str, writeable: bool) -> EncryptedFile:
        raise NotImplementedError("_get_file() not implemented.")

    def write_file(self, file: EncryptedFile, content: bytes):
        recipients = self._get_recipients_for_encryption()
        if not recipients:
            raise ValueError(
                "No recipients found for environment. "
                "Please add a 'batou.members' section to the secrets file."
            )
        file.write(content, recipients)

    def write_config_new(self, content: bytes):
        self.config_file.writeable = True
        with self.config_file:
            self.write_config(content)

    def _get_recipients(self) -> List[str]:
        recipients = self.config.get("batou", "members")

        if recipients.value is None:
            return []
        recipients = re.split(r"(\n|,)+", recipients.value)
        recipients = [r.strip() for r in recipients if r.strip()]
        return recipients


class GPGSecretProvider(ConfigFileSecretProvider):
    secret_provider_str = "gpg"

    def __init__(self, environment: "Environment"):
        super().__init__(environment)
        # load encrypted file
        self.config_file = GPGEncryptedFile(
            pathlib.Path(environment.base_dir)
            / "environments"
            / environment.name
            / "secrets.cfg.gpg"
        )

    def iter_secret_files(self, writeable=False) -> Dict[str, EncryptedFile]:
        environment_path = (
            pathlib.Path(self.environment.base_dir)
            / "environments"
            / self.environment.name
        )
        secret_files: Dict[str, EncryptedFile] = {}
        for file in environment_path.glob("secret-*.gpg"):
            file = environment_path / file
            if not file.is_file():
                continue
            name = file.name[len("secret-") : -len(".gpg")]
            secret_files[name] = GPGEncryptedFile(file, writeable)
        return secret_files

    def _get_file(self, name: str, writeable: bool = False) -> EncryptedFile:
        return GPGEncryptedFile(
            pathlib.Path(self.environment.base_dir)
            / "environments"
            / self.environment.name
            / f"secret-{name}.gpg",
            writeable,
        )

    def _get_recipients_for_encryption(self) -> List[str]:
        return self._get_recipients()

    def write_config(self, content: bytes, force_reencrypt: bool = False):
        config = ConfigUpdater().read_string(content.decode("utf-8"))
        secret_provider = config.get("batou", "secret_provider", fallback=None)
        if secret_provider is None or secret_provider.value is None:
            config.set("batou", "secret_provider", "gpg")
            print("Setting secret provider to gpg.")
        secret_provider = config.get("batou", "secret_provider")
        if secret_provider.value != "gpg":
            self.change_secret_provider(config, self)
            return
        recipients_opt = config.get("batou", "members")
        if recipients_opt is None or recipients_opt.value is None:
            raise ValueError(
                "Please add a 'batou.members' section to the secrets file."
            )
        recipients = re.split(r"(\n|,)+", recipients_opt.value)
        recipients = [r.strip() for r in recipients if r.strip()]
        if not recipients or len(recipients) == 0 or recipients[0] == "":
            raise ValueError(
                "Please add at least one recipient to the secrets file."
            )
        self.config_file.write(
            str(config).encode("utf-8"),
            recipients,
            reencrypt=force_reencrypt,
        )
        self.write_secret_files(self.read_secret_files())


def process_age_recipients(members, environment_path):
    """Process the recipients list."""
    key_meta_file_path = os.path.join(
        environment_path,
        "age_keys.txt",
    )
    key_meta_file_content = """\
###########################################
# This file is automatically generated by #
# batou. It contains the public keys of   #
# the members of the environment. It is   #
# re-written every time a secrets file is #
# encrypted (after a change).             #
###########################################
"""
    new_members = []
    for key in members:
        key = key.strip()
        if key.startswith("ssh-"):
            # it's a plain ssh public key and we can simply add it
            new_members.append(key)
            key_meta_file_content += f"# plain ssh public key\n{key}\n"
        elif key.startswith("age1"):
            # it's a plain age public key and we can simply add it
            new_members.append(key)
            key_meta_file_content += f"# plain age public key\n{key}\n"
        elif key.startswith("https://") or key.startswith("http://"):
            # it's a url to a key file, so we need to download it
            # and add it to the key meta file
            if key.startswith("http://"):
                raise ValueError(
                    "Downloading public keys over http is insecure!"
                )
            key_meta_file_content += f"# ssh key file from {key}\n"
            if debug:
                print(f"Downloading key file from `{key}`")
            key_file = urllib.request.urlopen(key)
            key_file_content = key_file.read().decode("utf-8")
            remote_keys = sorted(
                [
                    line
                    for line in key_file_content.splitlines()
                    if line.startswith("ssh-")
                ]
            )
            for line in remote_keys:
                new_members.append(line)
                key_meta_file_content += f"{line}\n"
        else:
            # unknown key type
            print(f"WARNING: Unknown key type for {key}\nWill be ignored!")
    # compare key_meta_file_content and the old one
    # if they differ, we will warn
    keys_changed = False
    if os.path.exists(key_meta_file_path):
        with open(key_meta_file_path, "r") as f:
            old_key_meta_file_content = f.read()
        if old_key_meta_file_content != key_meta_file_content:
            keys_changed = True
            print(
                "WARNING: The age encryption public-key metadata file has changed!\n"
                "This means that some secrets are now encrypted with a different set of keys.\n"
                "Please make sure that the new keys are correct and check the file in once you are done."
            )
    else:
        keys_changed = True
        print(
            "WARNING: The age encryption public-key metadata file does not exist!\n"
            "This is not a problem if you are setting up the environment for the first time.\n"
            "Please make sure that the new keys are correct and check the file in once you are done."
        )
    # write the new key meta file
    with open(key_meta_file_path, "w") as f:
        f.write(key_meta_file_content)
    return new_members, keys_changed


class AGESecretProvider(ConfigFileSecretProvider):
    FileFactory = AGEEncryptedFile
    secret_provider_str = "age"

    def __init__(self, environment: "Environment"):
        super().__init__(environment)
        self.config_file = self.FileFactory(
            pathlib.Path(environment.base_dir)
            / "environments"
            / environment.name
            / "secrets.cfg.age"
        )

    def iter_secret_files(self, writeable=False) -> Dict[str, EncryptedFile]:
        environment_path = (
            pathlib.Path(self.environment.base_dir)
            / "environments"
            / self.environment.name
        )
        secret_files: Dict[str, EncryptedFile] = {}
        for file in environment_path.glob("secret-*.age"):
            file = environment_path / file
            if not file.is_file():
                continue
            name = file.name[len("secret-") : -len(".age")]
            secret_files[name] = AGEEncryptedFile(file, writeable)
        return secret_files

    def _get_file(self, name: str, writeable: bool = False) -> EncryptedFile:
        return AGEEncryptedFile(
            pathlib.Path(self.environment.base_dir)
            / "environments"
            / self.environment.name
            / f"secret-{name}.age",
            writeable,
        )

    def _get_recipients_for_encryption(self) -> List[str]:
        recipients = self._get_recipients()
        return process_age_recipients(
            recipients,
            pathlib.Path(self.environment.base_dir)
            / pathlib.Path("environments")
            / self.environment.name,
        )[0]

    def write_config(self, content: bytes, force_reencrypt: bool = False):
        config = ConfigUpdater().read_string(content.decode("utf-8"))
        secret_provider = config.get("batou", "secret_provider", fallback=None)
        if secret_provider is None or secret_provider.value is None:
            config.set("batou", "secret_provider", self.secret_provider_str)
            print(f"Setting secret provider to {self.secret_provider_str}.")
        secret_provider = config.get("batou", "secret_provider")
        if secret_provider.value != self.secret_provider_str:
            self.change_secret_provider(config, self)
            return
        recipients_opt = config.get("batou", "members")
        if recipients_opt is None or recipients_opt.value is None:
            raise ValueError(
                "Please add a 'batou.members' section to the secrets file."
            )
        recipients = re.split(r"(\n|,)+", recipients_opt.value)
        recipients = [r.strip() for r in recipients if r.strip()]
        if not recipients or len(recipients) == 0 or recipients[0] == "":
            raise ValueError(
                "Please add at least one recipient to the secrets file."
            )
        recipients, keys_changed = process_age_recipients(
            recipients,
            pathlib.Path(self.environment.base_dir)
            / pathlib.Path("environments")
            / self.environment.name,
        )
        self.config_file.write(
            str(config).encode("utf-8"),
            recipients,
            reencrypt=keys_changed or force_reencrypt,
        )
        if keys_changed or force_reencrypt:
            self.write_secret_files(self.read_secret_files())


class DiffableAGESecretProvider(AGESecretProvider):
    FileFactory = DiffableAGEEncryptedFile
    secret_provider_str = "age-diffable"

    def __init__(self, environment: "Environment"):
        super().__init__(environment)
        self.config_file = DiffableAGEEncryptedFile(
            pathlib.Path(environment.base_dir)
            / "environments"
            / environment.name
            / "secrets.cfg.age-diffable"
        )
